import math
from itertools import cycle
from textwrap import wrap
from typing import Any

import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
from plotly.subplots import make_subplots

from pyBKTR.bktr import BKTRRegressor
from pyBKTR.utils import get_label_index_or_raise


def plot_temporal_betas(
    bktr_reg: BKTRRegressor,
    plot_feature_labels: list[str],
    spatial_point_label: str,
    show_figure: bool = True,
    fig_width: int = 850,
    fig_height: int = 550,
):
    """Create a plot of the beta values through time for a given spatial point and a set of
        feature labels.

    Args:
        bktr_reg (BKTRRegressor): BKTRRegressor object.
        plot_feature_labels (list[str]): List of feature labels to plot.
        spatial_point_label (str): Spatial point label to plot.
        show_figure (bool, optional): Whether to show the figure. Defaults to True.
        fig_width (int, optional): Figure width. Defaults to 850.
        fig_height (int, optional): Figure height. Defaults to 550.
    """
    if not bktr_reg.has_completed_sampling:
        raise RuntimeError('Plots can only be accessed after MCMC sampling.')
    # Verify all labels are valid
    get_label_index_or_raise(spatial_point_label, bktr_reg.spatial_labels, 'spatial')
    for feature_label in plot_feature_labels:
        get_label_index_or_raise(feature_label, bktr_reg.feature_labels, 'feature')

    scatters = []
    colorscale_hexas = px.colors.qualitative.Plotly
    color_cycle = cycle(colorscale_hexas)

    beta_summary_df: pd.DataFrame = bktr_reg.get_beta_summary_df(
        [spatial_point_label],
        None,
        plot_feature_labels,
    )
    for feature_label in plot_feature_labels:
        beta_df = beta_summary_df.loc[
            (spatial_point_label, slice(None), feature_label),
            ['Low2.5p', 'Mean', 'Up97.5p'],
        ]
        col_title = feature_label.replace('_', ' ').title()
        line_color = next(color_cycle)
        fill_rgba = _hex_to_rgba(line_color, 0.2)
        pctl_025 = beta_df['Low2.5p'].to_list()
        pctl_975 = beta_df['Up97.5p'].to_list()

        scatters.extend(
            [
                go.Scatter(
                    name=col_title,
                    x=bktr_reg.temporal_labels,
                    y=beta_df['Mean'],
                    mode='lines',
                    line=dict(color=line_color),
                ),
                go.Scatter(
                    name=f'{col_title} Bounds',
                    x=bktr_reg.temporal_labels + bktr_reg.temporal_labels[::-1],
                    y=pctl_975 + pctl_025[::-1],
                    mode='lines',
                    fillcolor=fill_rgba,
                    line=dict(width=0),
                    fill='toself',
                    hoverinfo='skip',
                    showlegend=False,
                ),
            ]
        )

    fig = go.Figure(scatters)
    fig.update_layout(
        yaxis_title='Beta Value',
        title=f'Location: {spatial_point_label.title()}',
        hovermode='x',
        width=fig_width,
        height=fig_height,
    )
    if show_figure:
        fig.show()
    return fig


def plot_spatial_betas(
    bktr_reg: BKTRRegressor,
    plot_feature_labels: list[str],
    temporal_point_label: str,
    nb_cols: int = 1,
    is_map: bool = True,
    mapbox_zoom: int = 9,
    use_dark_mode: bool = True,
    show_figure: bool = True,
    fig_width: int = 900,
    fig_height: int = 550,
):
    """Create a plot of beta values through space for a given temporal point and a set of
        feature labels.

    Args:
        bktr_reg (BKTRRegressor): BKTRRegressor object.
        plot_feature_labels (list[str]): List of feature labels to plot.
        temporal_point_label (str): Temporal point label to plot.
        nb_cols (int, optional): Number of columns in the plot. Defaults to 1.
        is_map (bool, optional): Whether to plot as a map. Defaults to True.
        mapbox_zoom (int, optional): Mapbox zoom. Defaults to 9. Only used if is_map is True.
        use_dark_mode (bool, optional): Whether to use dark mode. Defaults to True. Only used if
            is_map is True.
        show_figure (bool, optional): Whether to show the figure. Defaults to True.
        fig_width (int, optional): Figure width. Defaults to 900.
        fig_height (int, optional): Figure height. Defaults to 550.
    """
    if not bktr_reg.has_completed_sampling:
        raise RuntimeError('Plots can only be accessed after MCMC sampling.')
    # Verify all labels are valid
    get_label_index_or_raise(temporal_point_label, bktr_reg.temporal_labels, 'temporal')
    for feature_label in plot_feature_labels:
        get_label_index_or_raise(feature_label, bktr_reg.feature_labels, 'feature')

    beta_df = bktr_reg.beta_estimates.loc[(slice(None), temporal_point_label), plot_feature_labels]
    beta_df.melt(
        var_name='feature', value_vars=plot_feature_labels, ignore_index=False
    ).reset_index()

    feature_titles = [s.replace('_', ' ').title() for s in plot_feature_labels]
    min_beta, max_beta = beta_df.min().min(), beta_df.max().max()
    nb_subplots = len(plot_feature_labels)
    nb_rows = math.ceil(nb_subplots / nb_cols)
    coords_projector = bktr_reg.geo_coords_projector
    is_map = coords_projector is not None
    df_coord = coords_projector.ini_df.copy() if is_map else bktr_reg.spatial_positions_df.copy()
    if df_coord.shape[1] != 2:
        raise ValueError('Spatial coordinates must have 2 columns to be plotted.')
    if not is_map:
        plot_df = beta_df.melt(
            var_name='feature', value_vars=plot_feature_labels, ignore_index=False
        ).reset_index()
        plot_df = plot_df.merge(df_coord.reset_index(), on='location')
        col_x, col_y = df_coord.columns
        fig = px.scatter(
            plot_df,
            facet_col='feature',
            x=col_x,
            y=col_y,
            color='value',
            height=fig_height,
            width=fig_width,
        )
        if show_figure:
            fig.show()
            return
        return fig
    fig = make_subplots(
        rows=nb_rows,
        cols=nb_cols,
        subplot_titles=feature_titles,
        specs=[[{'type': 'mapbox'} for _ in range(nb_cols)] for _ in range(nb_rows)],
    )
    df_coord.columns = ['lat', 'lon']
    lat_list = df_coord['lat'].to_list()
    lon_list = df_coord['lon'].to_list()
    for i, feature_label in enumerate(plot_feature_labels):
        beta_df_feature = beta_df.loc[(slice(None), temporal_point_label), feature_label]
        beta_col_list = beta_df_feature.to_list()
        fig.add_trace(
            go.Scattermapbox(
                lat=lat_list,
                lon=lon_list,
                mode='markers',
                marker=go.scattermapbox.Marker(
                    cmin=min_beta,
                    cmax=max_beta,
                    color=beta_col_list,
                    size=7,
                    showscale=i == 0,
                    colorscale=px.colors.sequential.Viridis,
                    colorbar=dict(
                        title='Beta Values',
                        thickness=20,
                        titleside='top',
                        outlinecolor='rgba(68,68,68,0)',
                        ticks='outside',
                        ticklen=3,
                    ),
                ),
                text=[feature_titles],
            ),
            row=i // nb_cols + 1,
            col=i % nb_cols + 1,
        )
    fig.update_mapboxes(
        zoom=mapbox_zoom,
        style='carto-darkmatter' if use_dark_mode else 'carto-positron',
        center=go.layout.mapbox.Center(
            lat=sum(lat_list) / len(lat_list), lon=sum(lon_list) / len(lon_list)
        ),
    )
    fig.update_layout(
        showlegend=False,
        width=fig_width,
        height=fig_height,
    )
    if show_figure:
        fig.show()
        return
    return fig


def plot_beta_dists(
    bktr_reg: BKTRRegressor,
    labels_list: list[tuple[Any, Any, Any]],
    show_figure: bool = True,
    fig_width: int = 900,
    fig_height: int = 600,
):
    """Plot the distribution of beta values for a given list of labels.

    Args:
        bktr_reg (BKTRRegressor): BKTRRegressor object.
        labels_list (list[tuple[Any, Any, Any]]): List of labels (spatial, temporal, feature)
            for which to plot the beta distribution through iterations.
        show_figure (bool, optional): Whether to show the figure. Defaults to True.
        fig_width (int, optional): Figure width. Defaults to 900.
        fig_height (int, optional): Figure height. Defaults to 600.
    """
    if not bktr_reg.has_completed_sampling:
        raise RuntimeError('Plots can only be accessed after MCMC sampling.')
    betas_list = [bktr_reg.get_iterations_betas(*lab) for lab in labels_list]
    group_names = ['<br>'.join(lab) for lab in labels_list]
    fig = go.Figure()

    for i, betas in enumerate(betas_list):
        df = pd.DataFrame({'beta': betas, 'labels': [group_names[i]] * len(betas)})
        fig.add_trace(
            go.Violin(
                x=df['labels'],
                y=df['beta'],
                name=group_names[i],
                box_visible=True,
                meanline_visible=True,
            )
        )
    fig.update_layout(
        showlegend=False,
        width=fig_width,
        height=fig_height,
        xaxis={'type': 'category'},
        yaxis_title='Beta Value',
        title=(
            'Posterior distribution of beta values per given'
            ' spatial point, temporal point and feature'
        ),
    )
    if show_figure:
        fig.show()
        return
    return fig


def plot_covariates_beta_dists(
    bktr_reg: BKTRRegressor,
    feature_labels: list[Any] | None = None,
    show_figure: bool = True,
    fig_width: int = 900,
    fig_height: int = 600,
):
    """Plot the distribution of beta estimates regrouped by covariates.

    Args:
        bktr_reg (BKTRRegressor): BKTRRegressor object.
        feature_labels (list[Any] | None, optional): List of feature labels for
            which to plot the beta estimates distribution. If None plot for all features.
        show_figure (bool, optional): Whether to show the figure. Defaults to True.
        fig_width (int, optional): Figure width. Defaults to 900.
        fig_height (int, optional): Figure height. Defaults to 600.
    """
    if not bktr_reg.has_completed_sampling:
        raise RuntimeError('Plots can only be accessed after MCMC sampling.')
    if feature_labels is None:
        feature_labels = bktr_reg.feature_labels
    fig = go.Figure()
    for lab in feature_labels:
        if lab not in bktr_reg.beta_estimates.columns:
            raise ValueError(f'Invalid covariate label: {lab}')
        df = pd.DataFrame({'beta': bktr_reg.beta_estimates[[lab]][lab].to_list(), 'feature': lab})
        fig.add_trace(
            go.Violin(
                x=df['feature'],
                y=df['beta'],
                name=lab,
                box_visible=True,
                meanline_visible=True,
            )
        )
    fig.update_layout(
        showlegend=False,
        width=fig_width,
        height=fig_height,
        xaxis={'type': 'category'},
        yaxis_title='Beta Value',
        title='Distribution of beta estimates by feature across time and space',
    )
    if show_figure:
        fig.show()
        return
    return fig


def plot_hyperparams_dists(
    bktr_reg: BKTRRegressor,
    hyperparameters: list[str] | None = None,
    show_figure: bool = True,
    fig_width: int = 900,
    fig_height: int = 600,
):
    """Plot the distribution of hyperparameters through iterations.

    Args:
        hyperparameters (list[str] | None, optional): List of hyperparameters to plot.
            If None, plot all hyperparameters. Defaults to None.
        show_figure (bool, optional): Whether to show the figure. Defaults to True.
        fig_width (int, optional): Figure width. Defaults to 900.
        fig_height (int, optional): Figure height. Defaults to 600.
    """
    if not bktr_reg.has_completed_sampling:
        raise RuntimeError('Plots can only be accessed after MCMC sampling.')
    hparams_per_iter_df = bktr_reg.result_logger.hyperparameters_per_iter_df
    hparams = hparams_per_iter_df.columns if hyperparameters is None else hyperparameters
    fig = go.Figure()
    hparam_diff = set(hparams) - set(hparams_per_iter_df.columns)
    if hparam_diff:
        formatted_available_params = '\n'.join(hparams_per_iter_df.columns)
        formatted_hparam_diff = ', '.join(hparam_diff)
        raise ValueError(
            f'Hyperparameter(s) {formatted_hparam_diff} not found.'
            f' Available hyperparameters are:\n{formatted_available_params}'
        )
    for hparam in hparams:
        df = hparams_per_iter_df[[hparam]]
        fig.add_trace(
            go.Violin(
                x=[hparam] * len(df),
                y=df[hparam],
                name=hparam,
                box_visible=True,
                meanline_visible=True,
            )
        )
    fig.update_layout(
        showlegend=False,
        width=fig_width,
        height=fig_height,
        xaxis={'type': 'category'},
        yaxis_title='Hyperparameter Value',
        title='Posterior distribution of BKTR hyperparameters',
    )
    if show_figure:
        fig.show()
        return
    return fig


def plot_hyperparams_traceplot(
    bktr_reg: BKTRRegressor,
    hyperparameters: list[str] | None = None,
    show_figure: bool = True,
    fig_width: int = 800,
    fig_height: int = 550,
):
    """Plot the evolution of hyperparameters through iterations. (Traceplot)

    Args:
        hyperparameters (list[str] | None, optional): List of hyperparameters to plot.
            If None, plot all hyperparameters. Defaults to None.
        show_figure (bool, optional): Whether to show the figure. Defaults to True.
        fig_width (int, optional): Figure width. Defaults to 800.
        fig_height (int, optional): Figure height. Defaults to 550.
    """
    if not bktr_reg.has_completed_sampling:
        raise RuntimeError('Plots can only be accessed after MCMC sampling.')
    hparams_per_iter_df = bktr_reg.result_logger.hyperparameters_per_iter_df
    hparams = hparams_per_iter_df.columns if hyperparameters is None else hyperparameters
    hparam_diff = set(hparams) - set(hparams_per_iter_df.columns)
    if hparam_diff:
        formatted_available_params = '\n'.join(hparams_per_iter_df.columns)
        formatted_hparam_diff = ', '.join(hparam_diff)
        raise ValueError(
            f'Hyperparameter(s) {formatted_hparam_diff} not found.'
            f' Available hyperparameters are:\n{formatted_available_params}'
        )
    df = hparams_per_iter_df[hparams].copy()
    df.reset_index(inplace=True)
    df = df.melt(id_vars=['Sampling Iter'], var_name='Hyperparameter', value_name='Value')

    fig = px.line(df, x='Sampling Iter', y='Value', color='Hyperparameter')
    fig.update_layout(
        width=fig_width,
        height=fig_height,
        xaxis={'type': 'category'},
        yaxis_title='Hyperparameter Value',
        title='Hyperparameter values through sampling iterations (Traceplot)',
        legend={
            'orientation': 'h',
            'yanchor': 'bottom',
            'y': -0.3,
            'x': 1,
            'xanchor': 'right',
        },
    )
    if show_figure:
        fig.show()
        return
    return fig


def _hex_to_rgba(hex: str, transparency: float) -> str:
    col_hex = hex.lstrip('#')
    # Transform each hex in dec
    col_rgb = tuple(int(s, 16) for s in wrap(col_hex, 2))
    return f'rgba({col_rgb[0]}, {col_rgb[1]}, {col_rgb[2]}, {transparency})'



import math
import warnings
from typing import Callable

import numpy as np
import torch

from pyBKTR.kernels import Kernel, KernelParameter
from pyBKTR.tensor_ops import TSR


class KernelParamSampler:
    """Class for kernel's hyperparameter sampling

    The KernelParamSampler encapsulate all the behavior related to kernel
    hyperparameters' sampling

    """

    __slots__ = ('kernel', 'marginal_ll_eval_fn')

    kernel: Kernel
    """Kernel used for the hyperparameter sampling process"""
    marginal_ll_eval_fn: Callable

    def __init__(
        self,
        kernel: Kernel,
        marginal_ll_eval_fn: Callable,
    ):
        self.kernel = kernel
        self.marginal_ll_eval_fn = marginal_ll_eval_fn

    @staticmethod
    def initialize_theta_bounds(param: KernelParameter) -> tuple[float, float]:
        """Initialize sampling bounds according to current theta value and sampling scale"""
        theta_range = param.slice_sampling_scale * float(TSR.rand(1))
        theta_min = max(math.log(param.value) - theta_range, math.log(param.lower_bound))
        theta_max = min(theta_min + param.slice_sampling_scale, math.log(param.upper_bound))
        return theta_min, theta_max

    def _prior_fn(self, param: KernelParameter) -> float:
        """
        Prior likelihood function for a given hyperparameter value
        """
        # return -0.5 * self.config.hypr_precision_prior * (theta - self.config.hypr_mu_prior) ** 2
        return -0.5 * param.hparam_precision * (math.log(param.value)) ** 2

    @staticmethod
    def sample_rand_theta_value(theta_min, theta_max):
        """Sample a random theta value within the sampling bounds"""
        return theta_min + (theta_max - theta_min) * float(TSR.rand(1))

    def sample_param(self, param: KernelParameter):
        """The complete kernel hyperparameter sampling process"""
        theta_min, theta_max = self.initialize_theta_bounds(param)
        initial_theta = math.log(param.value)
        self.kernel.kernel_gen()
        initial_marginal_likelihood = self.marginal_ll_eval_fn() + self._prior_fn(param)
        density_threshold = float(TSR.rand(1))

        while True:
            new_theta = self.sample_rand_theta_value(theta_min, theta_max)
            param.value = math.exp(new_theta)
            self.kernel.kernel_gen()
            new_marginal_likelihood = self.marginal_ll_eval_fn() + self._prior_fn(param)

            marg_ll_diff = new_marginal_likelihood - initial_marginal_likelihood
            if np.exp(np.clip(marg_ll_diff, -709.78, 709.78)) > density_threshold:
                return param.value
            if new_theta < initial_theta:
                theta_min = new_theta
            else:
                theta_max = new_theta

    def sample(self):
        for param in self.kernel.parameters:
            if not param.is_fixed:
                self.sample_param(param)


def sample_norm_multivariate(
    mean_vec: torch.Tensor, precision_upper_tri: torch.Tensor
) -> torch.Tensor:
    """Sample a vector from a normal multivariate distribution

    Args:
        mean_vec (torch.Tensor): A vector of normal distribution means
        precision_upper_tri (torch.Tensor): An upper triangular matrix of
            precision between the distributions

    Returns:
        torch.Tensor: A sampled vector from the given normal multivariate information
    """
    return (
        torch.linalg.solve_triangular(
            precision_upper_tri, TSR.randn_like(mean_vec).unsqueeze(1), upper=True
        ).squeeze()
        + mean_vec
    )


def get_cov_decomp_chol(
    spatial_decomp: torch.Tensor,
    time_decomp: torch.Tensor,
    covs: torch.Tensor,
    rank_cp: int,
    omega: torch.Tensor,
    tau: float,
    y: torch.Tensor,
    wish_precision_tensor: torch.Tensor,
):
    y_masked = omega * y
    # TODO Merge some parts with marginal ll of spatial and temporal
    # get corresponding norm multivariate mean
    b = TSR.khatri_rao_prod(spatial_decomp, time_decomp).reshape(
        [spatial_decomp.shape[0], time_decomp.shape[0], rank_cp]
    )
    psi_c = torch.einsum('ijk,ijl->ijlk', (covs, b))
    psi_c_mask = psi_c * omega.unsqueeze(2).unsqueeze(3).expand_as(psi_c)
    psi_c_mask = psi_c_mask.permute([1, 0, 2, 3]).reshape(
        [psi_c.shape[0] * psi_c.shape[1], psi_c.shape[2] * psi_c.shape[3]]
    )
    inv_s = TSR.kronecker_prod(TSR.eye(rank_cp), wish_precision_tensor)
    lambda_c = tau * psi_c_mask.t().matmul(psi_c_mask) + inv_s
    chol_lc = torch.linalg.cholesky(lambda_c)
    cc = torch.linalg.solve(chol_lc, psi_c_mask.t().matmul(y_masked.t().flatten()))
    return {'chol_lc': chol_lc, 'cc': cc}


class TauSampler:
    """Sampler class for the Tau precision parameter of the BKTR Algorithm"""

    b_0: float
    a_tau: float

    def __init__(self, a_0: float, b_0: float, nb_observations: int):
        self.b_0 = b_0
        self.a_tau = a_0 + 0.5 * nb_observations

    def sample(self, total_sq_error):
        b_tau = self.b_0 + 0.5 * total_sq_error
        return torch.distributions.Gamma(self.a_tau.cpu(), b_tau).sample()


class PrecisionMatrixSampler:
    """Sample class to get new precision matrices from Wishart distributions"""

    nb_covariates: int
    wish_df: int
    wish_precision_tensor: torch.Tensor

    def __init__(self, nb_covariates: int, rank_cp: int):
        self.nb_covariates = nb_covariates
        self.wish_df = nb_covariates + rank_cp

    def sample(self, covs_decomp: torch.Tensor):
        w = covs_decomp.matmul(covs_decomp.t()) + TSR.eye(self.nb_covariates)
        # Warning of singular sample omitted since it is resampled anyway in the torch module
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            self.wish_precision_tensor = torch.distributions.Wishart(
                df=self.wish_df, precision_matrix=w
            ).sample()
        return self.wish_precision_tensor



from __future__ import annotations

import abc
import math
from enum import Enum
from functools import cached_property
from typing import Callable, Literal

import pandas as pd
import torch
from plotly import express as px

from pyBKTR.distances import get_euclidean_dist_tsr
from pyBKTR.tensor_ops import TSR
from pyBKTR.utils import log

DEFAULT_LBOUND = 1e-3
DEFAULT_UBOUND = 1e3


class KernelParameter:

    value: float
    """The hyperparameter mean's prior value (Paper -- :math:`\\phi`) or its fixed value"""
    is_fixed: bool = False
    """Says if the kernel parameter is fixed or not (if it is fixed, there is no sampling)"""
    lower_bound: float = DEFAULT_LBOUND
    """Hyperparameter's minimal value during sampling (Paper -- :math:`\\phi_{min}`)"""
    upper_bound: float = DEFAULT_UBOUND
    """Hyperparameter's maximal value during sampling (Paper -- :math:`\\phi_{max}`)"""
    slice_sampling_scale: float
    """The sampling range's amplitude (Paper -- :math:`\\rho`)"""
    hparam_precision: float
    """WHAT IS THAT? TODO"""
    kernel: Kernel | None
    """The kernel associated with the parameter (it is set at kernel instanciation)"""
    name: str
    """The parameter's name"""

    def __init__(
        self,
        value: float,
        is_fixed: bool = False,
        lower_bound: float = DEFAULT_LBOUND,
        upper_bound: float = DEFAULT_UBOUND,
        slice_sampling_scale: float = log(10),
        hparam_precision: float = 1.0,
    ):
        self.value = value
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.is_fixed = is_fixed
        self.slice_sampling_scale = slice_sampling_scale
        self.hparam_precision = hparam_precision

    def set_kernel(self, kernel: Kernel, param_name: str):
        self.kernel = kernel
        self.name = param_name
        self.kernel.parameters.append(self)

    @property
    def full_name(self) -> str:
        if self.kernel is None:
            return self.name
        return f'{self.kernel._name} - {self.name}'

    def __repr__(self):
        rep_attrs = [f'val={self.value}']
        if self.is_fixed:
            rep_attrs.append('is_fixed=True')
        if self.lower_bound != DEFAULT_LBOUND:
            rep_attrs.append(f'lbound={self.lower_bound}')
        if self.upper_bound != DEFAULT_UBOUND:
            rep_attrs.append(f'ubound={self.upper_bound}')
        return f'KernelParam({", ".join(rep_attrs)})'


class Kernel(abc.ABC):
    """
    Abstract Class Template for kernels

    :meta private:
    """

    kernel_variance: float
    jitter_value: float | None
    distance_matrix: torch.Tensor | None
    parameters: list[KernelParameter] = []
    covariance_matrix: torch.Tensor
    positions_df: pd.DataFrame | None
    has_dist_matrix: bool

    @property
    @abc.abstractmethod
    def _name(self) -> str:
        pass

    @abc.abstractmethod
    def _core_kernel_fn(self) -> torch.Tensor:
        pass

    def __init__(self, kernel_variance: float, jitter_value: float | None) -> None:
        self.parameters = []
        self.kernel_variance = kernel_variance
        self.jitter_value = jitter_value

    def add_jitter_to_kernel(self):
        if self.jitter_value == 0:
            return
        jitter_val = TSR.default_jitter if self.jitter_value is None else self.jitter_value
        self.covariance_matrix += jitter_val * TSR.eye(self.covariance_matrix.shape[0])

    def kernel_gen(self) -> torch.Tensor:
        if self.positions_df is None:
            raise RuntimeError('Set `positions_df` via `set_positions` before kernel evaluation.')
        self.covariance_matrix = self.kernel_variance * self._core_kernel_fn()
        self.add_jitter_to_kernel()
        return self.covariance_matrix

    def set_positions(self, position_df: pd.DataFrame):
        self.positions_df = position_df
        positions_tensor = TSR.tensor(position_df.to_numpy())
        if self.has_dist_matrix:
            self.distance_matrix = get_euclidean_dist_tsr(positions_tensor)

    def plot(self, show_figure: bool = True):
        fig = px.imshow(
            self.kernel_gen(),
            title=f'{self._name} Covariance Matrix Heatmap',
            color_continuous_scale=px.colors.sequential.Viridis,
        )
        axis_title = self.positions_df.index.name or 'x'
        fig.update_layout(
            xaxis_title=f'<b>{axis_title}</b>',
            yaxis_title=f"<b>{axis_title}</b>'",
            coloraxis_colorbar={'title': 'Covariance'},
            height=700,
            width=700,
        )
        if show_figure:
            fig.show()
            return
        return fig

    def __mul__(self, other) -> KernelMulComposed:
        return KernelMulComposed(self, other, f'({self._name} * {other._name})')

    def __add__(self, other) -> KernelAddComposed:
        return KernelAddComposed(self, other, f'({self._name} + {other._name})')


class KernelWhiteNoise(Kernel):
    """White Noise Kernel"""

    distance_matrix: None
    has_dist_matrix = False
    _name: str = 'White Noise Kernel'

    def __init__(
        self,
        kernel_variance: float = 1,
        jitter_value: float | None = None,
    ) -> None:
        super().__init__(kernel_variance, jitter_value)

    def _core_kernel_fn(self) -> torch.Tensor:
        return TSR.eye(len(self.positions_df))


class KernelSE(Kernel):
    """Squared Exponential Kernel"""

    lengthscale: KernelParameter
    distance_matrix: torch.Tensor
    has_dist_matrix = True
    _name: str = 'SE Kernel'

    def __init__(
        self,
        lengthscale: KernelParameter = KernelParameter(2),
        kernel_variance: float = 1,
        jitter_value: float | None = None,
    ) -> None:
        super().__init__(kernel_variance, jitter_value)
        self.lengthscale = lengthscale
        self.lengthscale.set_kernel(self, 'lengthscale')

    def _core_kernel_fn(self) -> torch.Tensor:
        return torch.exp(-self.distance_matrix**2 / (2 * self.lengthscale.value**2))


class KernelRQ(Kernel):
    """Rational Quadratic Kernel"""

    lengthscale: KernelParameter
    alpha: KernelParameter
    distance_matrix: torch.Tensor
    has_dist_matrix = True
    _name: str = 'RQ Kernel'

    def __init__(
        self,
        lengthscale: KernelParameter = KernelParameter(2),
        alpha: KernelParameter = KernelParameter(2),
        kernel_variance: float = 1,
        jitter_value: float | None = None,
    ) -> None:
        super().__init__(kernel_variance, jitter_value)
        self.lengthscale = lengthscale
        self.lengthscale.set_kernel(self, 'lengthscale')
        self.alpha = alpha
        self.alpha.set_kernel(self, 'alpha')

    def _core_kernel_fn(self) -> torch.Tensor:
        return (
            1 + self.distance_matrix**2 / (2 * self.alpha.value * self.lengthscale.value**2)
        ) ** -self.alpha.value


class KernelPeriodic(Kernel):
    """Periodic Kernel"""

    lengthscale: KernelParameter
    period_length: KernelParameter
    distance_matrix: torch.Tensor
    has_dist_matrix = True
    _name: str = 'Periodic Kernel'

    def __init__(
        self,
        lengthscale: KernelParameter = KernelParameter(2),
        period_length: KernelParameter = KernelParameter(2),
        kernel_variance: float = 1,
        jitter_value: float | None = None,
    ) -> None:
        super().__init__(kernel_variance, jitter_value)
        self.lengthscale = lengthscale
        self.lengthscale.set_kernel(self, 'lengthscale')
        self.period_length = period_length
        self.period_length.set_kernel(self, 'period length')

    def _core_kernel_fn(self) -> torch.Tensor:
        return torch.exp(
            -2
            * torch.sin(torch.pi * self.distance_matrix / self.period_length.value) ** 2
            / self.lengthscale.value**2
        )


class KernelMatern(Kernel):
    lengthscale: KernelParameter
    smoothness_factor: Literal[1, 3, 5]
    distance_matrix: torch.Tensor
    has_dist_matrix = True
    _name: str = 'Matern Kernel'

    def __init__(
        self,
        smoothness_factor: Literal[1, 3, 5],
        lengthscale: KernelParameter = KernelParameter(2),
        kernel_variance: float = 1,
        jitter_value: float | None = None,
    ) -> None:
        if smoothness_factor not in {1, 3, 5}:
            raise ValueError('smoothness factor should be one of the following values 1, 3 or 5')
        super().__init__(kernel_variance, jitter_value)
        self._name = f'Matern {smoothness_factor}/2 Kernel'
        self.smoothness_factor = smoothness_factor
        self.lengthscale = lengthscale
        self.lengthscale.set_kernel(self, 'lengthscale')

    @cached_property
    def smoothness_kernel_fn(self) -> Callable:
        """
        Get the right kernel function associated with the right smoothness factor
        """
        match self.smoothness_factor:
            case 1:
                return lambda t: 1
            case 3:
                return lambda t: 1 + t
            case 5:
                return lambda t: 1 + t * (1 + t / 3)
            case _:
                raise ValueError('Kernel function for this smoothness factor is not implemented')

    def _core_kernel_fn(self) -> torch.Tensor:
        temp_kernel = (
            self.distance_matrix * math.sqrt(self.smoothness_factor) / self.lengthscale.value
        )
        return self.smoothness_kernel_fn(temp_kernel) * (-temp_kernel).exp()


class CompositionOps(Enum):
    """Type of arithmetic operation to be performed to compose a kernel

    :meta private:
    """

    MUL = 'mul'
    ADD = 'add'


class KernelComposed(Kernel):
    """
    Kernel composed of two kernels and created when two kernels are
    added or multiplied.
    """

    _name: str = ''
    parameters: list = []
    left_kernel = Kernel
    right_kernel = Kernel
    has_dist_matrix = True

    def __init__(
        self,
        left_kernel: Kernel,
        right_kernel: Kernel,
        new_name: str,
        composition_operation: CompositionOps,
    ) -> None:
        composed_variance = 1
        new_jitter_val = max(
            left_kernel.jitter_value or TSR.default_jitter,
            right_kernel.jitter_value or TSR.default_jitter,
        )
        super().__init__(composed_variance, new_jitter_val)
        self.left_kernel = left_kernel
        self.right_kernel = right_kernel
        self._name = new_name
        self.parameters = self.left_kernel.parameters + self.right_kernel.parameters
        self.composition_operation = composition_operation

    def _core_kernel_fn(self) -> torch.Tensor:
        match self.composition_operation:
            case CompositionOps.ADD:
                return self.left_kernel._core_kernel_fn() + self.right_kernel._core_kernel_fn()
            case CompositionOps.MUL:
                return self.left_kernel._core_kernel_fn() * self.right_kernel._core_kernel_fn()
        raise RuntimeError('Composition operation not implemented')

    def set_positions(self, positions_df: pd.DataFrame) -> None:
        super().set_positions(positions_df)
        self.left_kernel.set_positions(positions_df)
        self.right_kernel.set_positions(positions_df)


class KernelAddComposed(KernelComposed):
    """Kernel composed of two kernels and created when two kernels are added."""

    def __init__(self, left_kernel: Kernel, right_kernel: Kernel, new_name: str) -> None:
        super().__init__(left_kernel, right_kernel, new_name, CompositionOps.ADD)


class KernelMulComposed(KernelComposed):
    """Kernel composed of two kernels and created when two kernels are multiplied."""

    def __init__(self, left_kernel: Kernel, right_kernel: Kernel, new_name: str) -> None:
        super().__init__(left_kernel, right_kernel, new_name, CompositionOps.MUL)



import textwrap
from collections import defaultdict
from time import time
from typing import Any, Literal

import pandas as pd
import torch
from formulaic import Formula

from pyBKTR.kernels import Kernel, KernelAddComposed, KernelMulComposed
from pyBKTR.tensor_ops import TSR
from pyBKTR.utils import get_label_index_or_raise


class ResultLogger:
    __slots__ = [
        'y',
        'omega',
        'covariates',
        'nb_burn_in_iter',
        'nb_sampling_iter',
        'logged_params_map',
        'beta_estimates',
        'y_estimates',
        'total_elapsed_time',
        'formula',
        'rank_decomp',
        'spatial_labels',
        'temporal_labels',
        'feature_labels',
        'spatial_kernel',
        'temporal_kernel',
        'hparam_labels',
        'hparam_per_iter',
        'spatial_decomp_per_iter',
        'temporal_decomp_per_iter',
        'covs_decomp_per_iter',
        'sum_beta_est',
        'sum_y_est',
        'beta_estimates_df',
        'y_estimates_df',
        'beta_covariates_summary_df',
        'hyperparameters_per_iter_df',
        'last_time_stamp',
        'error_metrics',
        'total_sq_error',
    ]

    # Metrics used to create beta summaries
    moment_metrics = ['Mean', 'SD']
    quantile_metrics = [
        'Min',
        'Q1',
        'Median',
        'Q3',
        'Max',
        'Low2.5p',
        'Up97.5p',
    ]
    quantile_values = [0, 0.25, 0.5, 0.75, 1, 0.025, 0.975]

    # Summary parameters
    LINE_NCHAR = 70
    TAB_STR = '  '
    COL_WIDTH = 18
    DF_DISTRIB_STR_PARAMS = {
        'float_format': '{:,.3f}'.format,
        'col_space': 8,
        'line_width': LINE_NCHAR,
        'max_colwidth': COL_WIDTH,
        'formatters': {'__index__': lambda x, w=COL_WIDTH: f'{x:{w}}'},
    }
    MAIN_SUMMARY_COLS = ['Mean', 'Median', 'SD']
    DISTRIB_COLS = [*MAIN_SUMMARY_COLS, 'Low2.5p', 'Up97.5p']

    def __init__(
        self,
        y: torch.Tensor,
        omega: torch.Tensor,
        covariates: torch.Tensor,
        nb_burn_in_iter: int,
        nb_sampling_iter: int,
        rank_decomp: int,
        formula: Formula,
        spatial_labels: list,
        temporal_labels: list,
        feature_labels: list,
        spatial_kernel: Kernel,
        temporal_kernel: Kernel,
    ):
        # Create a tensor dictionary holding scalar data gathered through all iterations
        self.logged_params_map = defaultdict(list)

        # Create tensors that accumulate values needed for estimates
        self.spatial_labels = spatial_labels
        self.temporal_labels = temporal_labels
        self.feature_labels = feature_labels
        self.hparam_labels = [
            'Tau',
            *[f'Spatial - {p.full_name}' for p in spatial_kernel.parameters if not p.is_fixed],
            *[f'Temporal - {p.full_name}' for p in temporal_kernel.parameters if not p.is_fixed],
        ]
        self.spatial_decomp_per_iter = TSR.zeros(
            (len(spatial_labels), rank_decomp, nb_sampling_iter)
        )
        self.temporal_decomp_per_iter = TSR.zeros(
            (len(temporal_labels), rank_decomp, nb_sampling_iter)
        )
        self.covs_decomp_per_iter = TSR.zeros(
            (len(self.feature_labels), rank_decomp, nb_sampling_iter)
        )
        self.hparam_per_iter = TSR.zeros((len(self.hparam_labels), nb_sampling_iter))
        self.sum_beta_est = TSR.zeros(covariates.shape)
        self.sum_y_est = TSR.zeros(y.shape)
        self.total_elapsed_time = 0

        self.y = y
        self.omega = omega
        self.covariates = covariates
        self.formula = formula
        self.rank_decomp = rank_decomp
        self.spatial_kernel = spatial_kernel
        self.temporal_kernel = temporal_kernel
        self.nb_burn_in_iter = nb_burn_in_iter
        self.nb_sampling_iter = nb_sampling_iter

        # Set initial timer value to calculate iterations' processing time
        self.last_time_stamp = time()

    def collect_iter_samples(self, iter: int, tau_value: float):
        """
        Collect current iteration values inside the historical data tensor list.
        To be noted that errors have already been calculated before tau sampling.
        """
        elapsed_time = self._get_elapsed_time()

        if iter > self.nb_burn_in_iter:
            self.sum_beta_est += self.beta_estimates
            self.sum_y_est += self.y_estimates

            # Collect Hyperparameters
            s_iter = iter - self.nb_burn_in_iter - 1
            s_params = [p for p in self.spatial_kernel.parameters if not p.is_fixed]
            t_params = [p for p in self.temporal_kernel.parameters if not p.is_fixed]
            self.hparam_per_iter[0, s_iter] = tau_value
            self.hparam_per_iter[1 : len(s_params) + 1, s_iter] = TSR.tensor(
                [p.value for p in s_params]
            )
            self.hparam_per_iter[1 + len(s_params) :, s_iter] = TSR.tensor(
                [p.value for p in t_params]
            )

        total_logged_params = {
            **{
                'iter': iter,
                'is_burn_in': 1 if iter <= self.nb_burn_in_iter else 0,
                'elapsed_time': elapsed_time,
            },
            **self.error_metrics,
        }

        for k, v in total_logged_params.items():
            self.logged_params_map[k].append(v)

        self._print_iter_result(iter, elapsed_time)

    def _get_elapsed_time(self):
        iter_elapsed_time = time() - self.last_time_stamp
        self.total_elapsed_time += iter_elapsed_time
        self.last_time_stamp = time()
        return iter_elapsed_time

    def set_error_metrics(self):
        """Calculate error metrics of interest (MAE, RMSE, Total Error)"""
        nb_observ = self.omega.sum()
        err_matrix = (self.y_estimates - self.y) * self.omega
        total_sq_error = err_matrix.norm() ** 2
        mae = err_matrix.abs().sum() / nb_observ
        rmse = (total_sq_error / nb_observ).sqrt()
        self.total_sq_error = float(total_sq_error.cpu())
        self.error_metrics = {
            'MAE': float(mae.cpu()),
            'RMSE': float(rmse.cpu()),
        }

    def set_y_and_beta_estimates(self, decomp_tensors_map: dict[str, torch.Tensor], iter: int):
        """Calculate the estimated y and beta tensors

        Args:
            decomposition_tensors_map (dict[str, torch.Tensor]): A dictionnary that
                contains all the spatial, temporal and covariates decomposition
            iter (int): The current iteration number.
        """
        if iter > self.nb_burn_in_iter:
            iter_indx = iter - self.nb_burn_in_iter - 1
            self.spatial_decomp_per_iter[:, :, iter_indx] = decomp_tensors_map['spatial_decomp']
            self.temporal_decomp_per_iter[:, :, iter_indx] = decomp_tensors_map['temporal_decomp']
            self.covs_decomp_per_iter[:, :, iter_indx] = decomp_tensors_map['covs_decomp']

        self.beta_estimates = torch.einsum(
            'im,jm,km->ijk',
            [
                decomp_tensors_map['spatial_decomp'],
                decomp_tensors_map['temporal_decomp'],
                decomp_tensors_map['covs_decomp'],
            ],
        )
        self.y_estimates = torch.einsum('ijk,ijk->ij', self.covariates, self.beta_estimates)

    def _print_iter_result(self, iter: int | str, elapsed_time: float):
        formatted_errors = [f'{k} {v:7.4f}' for k, v in self.error_metrics.items()]
        result_items = [
            f'Iter {iter:<5}',
            f'Elapsed {elapsed_time:8.2f}s',
            *formatted_errors,
        ]
        print(f'* {" | ".join(result_items)} *')

    def log_final_iter_results(self):
        self.beta_estimates = self.sum_beta_est / self.nb_sampling_iter
        self.y_estimates = self.sum_y_est / self.nb_sampling_iter
        beta_covariates_summary = self._create_distrib_values_summary(
            self.beta_estimates.reshape(-1, len(self.feature_labels)), dim=0
        )
        self.beta_covariates_summary_df = pd.DataFrame(
            beta_covariates_summary.t().cpu(),
            index=pd.Index(self.feature_labels, name='feature'),
            columns=self.moment_metrics + self.quantile_metrics,
        )
        y_beta_index = pd.MultiIndex.from_product(
            [self.spatial_labels, self.temporal_labels], names=['location', 'time']
        )
        self.y_estimates_df = pd.DataFrame(
            self.y_estimates.flatten().cpu(),
            columns=[str(self.formula.lhs)],
            index=y_beta_index,
        )
        self.beta_estimates_df = pd.DataFrame(
            self.beta_estimates.reshape([-1, len(self.feature_labels)]).cpu(),
            columns=self.feature_labels,
            index=y_beta_index,
        )
        self.hyperparameters_per_iter_df = pd.DataFrame(
            self.hparam_per_iter.t().cpu(),
            columns=self.hparam_labels,
            index=pd.Index(range(1, self.nb_sampling_iter + 1), name='Sampling Iter'),
        )
        self.set_error_metrics()
        self._print_iter_result('TOTAL', self.total_elapsed_time)

    @classmethod
    def _create_distrib_values_summary(cls, values: torch.Tensor, dim: int = None) -> torch.Tensor:
        """Create a summary for a given tensor of beta values across a given dimension
        for the metrics set in the class.

        Args:
            values (torch.Tensor): Values to summarize
            dim (int, optional): Dimension of the tensor we want to summarize. If None,
                we want to summarize the whole tensor and flatten it. Defaults to None.

        Returns:
            torch.Tensor: A tensor with summaries for the given beta values
        """
        all_metrics = cls.moment_metrics + cls.quantile_metrics
        summary_shape = [len(all_metrics)]
        if dim is not None:
            beta_val_shape = list(values.shape)
            summary_shape = summary_shape + beta_val_shape[:dim] + beta_val_shape[dim + 1 :]
        beta_summaries = TSR.zeros(summary_shape)
        # In the advent of having no values, we return the empty tensor
        if values.numel() == 0:
            return beta_summaries
        # Dimension for moment calculations are a bit different than for quantile
        moment_dim = dim if dim is not None else []
        beta_summaries[0] = values.mean(dim=moment_dim)
        beta_summaries[1] = values.std(dim=moment_dim)
        beta_summaries[len(cls.moment_metrics) :] = torch.quantile(
            values, TSR.tensor(cls.quantile_values), dim=dim
        )
        return beta_summaries

    def get_beta_summary_df(
        self,
        spatial_labels: list[Any] = None,
        temporal_labels: list[Any] = None,
        feature_labels: list[Any] = None,
    ) -> pd.DataFrame:
        """Get a summary of the beta values. If no labels are given, then the summary is for all
            the betas. If labels are given, then the summary is for the given labels.

        Args:
            spatial_labels (list[Any], optional): The spatial labels to get the summary for.
                Defaults to None.
            temporal_labels (list[Any], optional): The temporal labels to get the summary for.
                Defaults to None.
            feature_labels (list[Any], optional): The feature labels to get the summary for.
                Defaults to None.

        Returns:
            pd.DataFrame: A dataframe with the summary for the given labels.
        """
        spatial_labs = spatial_labels if spatial_labels is not None else self.spatial_labels
        temporal_labs = temporal_labels if temporal_labels is not None else self.temporal_labels
        feature_labs = feature_labels if feature_labels is not None else self.feature_labels
        iteration_betas = self.get_iteration_betas_tensor(
            spatial_labs, temporal_labs, feature_labs
        )
        beta_summary = self._create_distrib_values_summary(iteration_betas, dim=1).t()
        return pd.DataFrame(
            beta_summary.cpu(),
            columns=self.moment_metrics + self.quantile_metrics,
            index=pd.MultiIndex.from_product([spatial_labs, temporal_labs, feature_labs]),
        )

    def get_iteration_betas_tensor(
        self, spatial_labels: list[Any], temporal_labels: list[Any], feature_labels: list[Any]
    ) -> torch.Tensor:
        spatial_indexes = _get_label_indexes(spatial_labels, self.spatial_labels, 'spatial')
        temporal_indexes = _get_label_indexes(temporal_labels, self.temporal_labels, 'temporal')
        feature_indexes = _get_label_indexes(feature_labels, self.feature_labels, 'feature')
        betas_per_iterations = torch.einsum(
            'sri,tri,cri->stci',
            [
                self.spatial_decomp_per_iter[spatial_indexes, :, :],
                self.temporal_decomp_per_iter[temporal_indexes, :, :],
                self.covs_decomp_per_iter[feature_indexes, :, :],
            ],
        )
        return betas_per_iterations.reshape([-1, self.nb_sampling_iter])

    def summary(self) -> str:
        """Print a summary of the BKTR regressor.

        Returns:
            str: A string representation of the BKTR regressor after MCMC sampling.
        """
        line_sep = self.LINE_NCHAR * '='
        title_format = f'^{self.LINE_NCHAR}'
        summary_str = [
            '',
            line_sep,
            f'{"BKTR Regressor Summary":{title_format}}',
            line_sep,
            self._get_formula_str(),
            '',
            f'Burn-in iterations: {self.nb_burn_in_iter}',
            f'Sampling iterations: {self.nb_sampling_iter}',
            f'Rank decomposition: {self.rank_decomp}',
            f'Nb Spatial Locations: {len(self.spatial_labels)}',
            f'Nb Temporal Points: {len(self.temporal_labels)}',
            f'Nb Covariates: {len(self.feature_labels)}',
            line_sep,
            'In Sample Errors:',
            f'{self.TAB_STR}RMSE: {self.error_metrics["RMSE"]:.3f}',
            f'{self.TAB_STR}MAE: {self.error_metrics["MAE"]:.3f}',
            f'Computation time: {self.total_elapsed_time:.2f}s.',
            line_sep,
            '-- Spatial Kernel --',
            self._kernel_summary(self.spatial_kernel, 'spatial'),
            '',
            '-- Temporal Kernel --',
            self._kernel_summary(self.temporal_kernel, 'temporal'),
            line_sep,
            self._beta_summary(),
            line_sep,
            '',
        ]
        return '\n'.join(summary_str)

    def _get_formula_str(self) -> str:
        formula_str = f'Formula: {self.formula.lhs} ~ {self.formula.rhs}'
        f_wrap = textwrap.wrap(formula_str, width=self.LINE_NCHAR, subsequent_indent=self.TAB_STR)
        return '\n'.join(f_wrap)

    def _kernel_summary(
        self, kernel: Kernel, kernel_type: Literal['spatial', 'temporal'], indent_count=0
    ) -> str:
        """Get a string representation of a given kernel. Since the kernel can be composed, this
            function needs to be recursive.


        Args:
            kernel (Kernel): The kernel to get the summary for.
            kernel_type (Literal[&#39;spatial&#39;, &#39;temporal&#39;]): The type of kernel.
            indent_count (int, optional): Indentation level (related to the depth of composition).
                Defaults to 0.

        Returns:
            str: A string representation of the kernel. Containing the name of the kernel,
                the estimated parameters distribution and the fixed parameters.
        """
        params = kernel.parameters
        if kernel.__class__ in [KernelAddComposed, KernelMulComposed]:
            new_ind_nb = indent_count + 1
            kernel_elems = [
                f'Composed Kernel ({str(kernel.composition_operation.value).capitalize()})',
                f'{self._kernel_summary(kernel.left_kernel, kernel_type, new_ind_nb)}',
                f'{self.TAB_STR}{"+" if kernel.composition_operation == "add" else "*"}',
                f'{self._kernel_summary(kernel.right_kernel, kernel_type, new_ind_nb)}',
            ]
        else:
            fixed_params = [p for p in params if p.is_fixed]
            sampled_params = [p for p in params if not p.is_fixed]
            sampled_par_indexes = [
                self.hparam_labels.index(f'{kernel_type.capitalize()} - {p.full_name}')
                for p in sampled_params
            ]
            sampled_par_tsr = self.hparam_per_iter[sampled_par_indexes, :]
            sampled_par_summary = self._create_distrib_values_summary(sampled_par_tsr, dim=1)
            sampled_par_df = pd.DataFrame(
                sampled_par_summary.cpu().t(),
                columns=self.moment_metrics + self.quantile_metrics,
                index=[p.name for p in sampled_params],
            )[self.DISTRIB_COLS]
            sampled_params_str = sampled_par_df.to_string(**self.DF_DISTRIB_STR_PARAMS)
            constant_params_strs = [
                f'{p.name:20}   Fixed Value: {p.value:.3f}' for p in fixed_params
            ]
            kernel_elems = [
                kernel._name,
                'Parameter(s):',
                sampled_params_str,
                *constant_params_strs,
            ]
        kernel_str = '\n'.join(kernel_elems)
        return textwrap.indent(kernel_str, self.TAB_STR * indent_count)

    def _beta_summary(self) -> str:
        """Get a string representation of the beta estimates aggregated per covariates.
            (This shows the distribution of the beta hats per covariates)

        Returns:
            str: A string representation of the beta estimates.
        """

        distrib_df = self.beta_covariates_summary_df[self.MAIN_SUMMARY_COLS].copy()
        beta_distrib_str = distrib_df.to_string(**self.DF_DISTRIB_STR_PARAMS)
        beta_summary_str = [
            'Beta Estimates Summary (Aggregated Per Covariates)',
            '',
            beta_distrib_str,
        ]
        return '\n'.join(beta_summary_str)


def _get_label_indexes(
    labels: list[Any],
    available_labels: list[Any],
    label_type: Literal['spatial', 'temporal', 'feature'],
) -> list[int]:
    """return the indexes of a given set of labels that can be found in a list of available labels

    Args:
        labels (list[Any]): The labels for which we want to get the indexes
        available_labels (list[Any]): The list of available labels
        label_type (Literal[&#39;spatial&#39;, &#39;temporal&#39;, &#39;feature&#39;]): Type
            of label for which we want to get indexes

    Raises:
        ValueError: Error if the list of labels is empty

    Returns:
        list[int]: The indexes of the labels in the list of available labels
    """
    if len(labels) == 0:
        raise ValueError(f'No {label_type} labels provided.')
    return [get_label_index_or_raise(lab, available_labels, label_type) for lab in labels]



import numpy as np
import pandas as pd
from pkg_resources import resource_stream

from pyBKTR.utils import reshape_covariate_dfs


class BixiData:
    departure_df: pd.DataFrame
    temporal_features_df: pd.DataFrame
    spatial_features_df: pd.DataFrame
    spatial_positions_df: pd.DataFrame
    temporal_positions_df: pd.DataFrame
    data_df: pd.DataFrame

    def __init__(self):
        self.departure_df = self._get_source_df('bixi_station_departures')
        # Normalize departure counts
        max_val = np.max(self.departure_df.to_numpy(na_value=0))
        self.departure_df = self.departure_df / max_val

        w_df = self._get_source_df('bixi_temporal_features')
        # Normalize each column of the weather data
        self.temporal_features_df = (w_df - w_df.min()) / (w_df.max() - w_df.min())

        s_df = self._get_source_df('bixi_spatial_features')
        # Normalize each column of the station data
        self.spatial_features_df = (s_df - s_df.min()) / (s_df.max() - s_df.min())

        self.spatial_positions_df = self._get_source_df('bixi_spatial_locations')
        self.temporal_positions_df = self._get_source_df('bixi_temporal_locations')
        self.data_df = reshape_covariate_dfs(
            spatial_df=self.spatial_features_df,
            temporal_df=self.temporal_features_df,
            y_df=self.departure_df,
            y_column_name='nb_departure',
        )

    @staticmethod
    def _get_source_df(csv_name: str) -> pd.DataFrame:
        file_name = resource_stream('pyBKTR', f'data/{csv_name}.csv')
        return pd.read_csv(file_name, index_col=0)



import torch

from pyBKTR.tensor_ops import TSR


class MarginalLikelihoodEvaluator:
    """Class that evaluate the marginal likelihood of the kernel hyperparameter

    MarginalLikelihoodEvaluator enable the calculation of the marginal
    likelihood of the kernel hyperparameters. This likelihood is used during the sampling
    process.

    """

    axis_permutation: list[int]
    rank_decomp: int
    nb_covariates: int
    covariates: torch.Tensor
    omega: torch.Tensor
    y_masked: torch.Tensor

    inv_k: torch.Tensor
    chol_k: torch.Tensor
    chol_lu: torch.Tensor
    # TODO What is uu
    uu: torch.Tensor
    likelihood: float

    def __init__(
        self,
        rank_decomp: int,
        nb_covariates: int,
        covariates: torch.Tensor,
        omega: torch.Tensor,
        y: torch.Tensor,
        is_transposed: bool,
    ):
        self.rank_decomp = rank_decomp
        self.nb_covariates = nb_covariates
        self.covariates = covariates
        self.omega = omega
        self.axis_permutation = [1, 0] if is_transposed else [0, 1]
        self.y_masked = y * omega

    def calc_likelihood(
        self,
        kernel_values: torch.Tensor,
        decomp_values: torch.Tensor,
        covs_decomp: torch.Tensor,
        tau: float,
    ) -> float:
        """Calculate and set the marginal likelihood according to current parameter values

        Args:
            kernel_values (torch.Tensor): Current values of the related kernel for the
            decomp_values (torch.Tensor): Values of the related tensor decomposition
            covs_decomp (torch.Tensor): Complete tensor of the covariates
            tau (float): current value of the white noise precision

        Returns:
            float: The calculated marginal likelihood
        """
        rank_decomp = self.rank_decomp
        kernel_size = kernel_values.shape[0]
        lambda_size = kernel_size * self.rank_decomp

        psi_u = torch.einsum(
            'ijk,jkl->ilj',
            [
                self.covariates.permute([*self.axis_permutation, 2]),
                TSR.khatri_rao_prod(decomp_values, covs_decomp).reshape(
                    [-1, self.nb_covariates, rank_decomp]
                ),
            ],
        )
        psi_u_mask = psi_u * (
            self.omega.permute(self.axis_permutation).unsqueeze(1).expand_as(psi_u)
        )

        self.chol_k = torch.linalg.cholesky(kernel_values)
        kernel_inverse = torch.linalg.solve(
            self.chol_k.t(), torch.linalg.solve(self.chol_k, TSR.eye(kernel_size))
        )
        stabilized_kernel_inv = (kernel_inverse.t() + kernel_inverse) / 2
        self.inv_k = TSR.kronecker_prod(
            TSR.eye(rank_decomp), stabilized_kernel_inv
        )  # I_R Kron inv(Ks)

        lambda_u = tau * torch.einsum('ijk,ilk->ijl', [psi_u_mask, psi_u_mask])  # tau * H_T * H_T'
        lambda_u = (
            (lambda_u.transpose(0, -1).unsqueeze(-1) * TSR.eye(kernel_size))
            .transpose(1, 2)
            .reshape(lambda_size, lambda_size)
        )
        lambda_u += self.inv_k

        self.chol_lu = torch.linalg.cholesky(lambda_u)
        self.uu = torch.linalg.solve_triangular(
            self.chol_lu,
            torch.einsum('ijk,ik->ji', [psi_u_mask, self.y_masked.permute(self.axis_permutation)])
            .flatten()
            .unsqueeze(1),
            upper=False,
        ).squeeze()
        self.likelihood = float(
            (
                0.5 * tau**2 * self.uu.t().matmul(self.uu)
                - self.chol_lu.diag().log().sum()
                - TSR.tensor(rank_decomp) * self.chol_k.diag().log().sum()
            ).cpu()
        )

        return self.likelihood



from datetime import datetime

import numpy as np
import pandas as pd
from pkg_resources import resource_stream


def get_source_file_name(csv_name: str) -> str:
    return resource_stream('pyBKTR', f'data/raw_data/{csv_name}.csv')


EXPORT_PATH = '/Users/juju/Projects/pyBKTR/pyBKTR/data'
date_columns = ['day', 'month', 'weekday']

# Temporal features
df = pd.read_csv(get_source_file_name('montreal_weather_data'))
df.index = (
    df['day']
    .apply(lambda x: datetime.strptime(f'{2019}-{x}', '%Y-%j').strftime('%Y-%m-%d'))
    .to_list()
)
df = df.drop(columns=date_columns)
df.index.name = 'time'
df = df[1:197]
df = df.rename(
    columns={
        'Mean Temp (°C)': 'mean_temp_c',
        'Total Precip (mm)': 'total_precip_mm',
        'max_temp': 'max_temp_f',
    }
)
df.to_csv(f'{EXPORT_PATH}/bixi_temporal_features.csv')

# Temporal Locations
df['time_index'] = range(0, len(df))
df = df[['time_index']]
df.to_csv(f'{EXPORT_PATH}/bixi_temporal_locations.csv')


# Spatial features
df = pd.read_csv(get_source_file_name('bike_station_features'))
df.index = df['Code'].astype('str') + ' - ' + df['name']
df.index.name = 'location'
s_df = df
df = df[df.columns[8:21]]
df.to_csv(f'{EXPORT_PATH}/bixi_spatial_features.csv')

# Spatial Locations
df = s_df[['latitude', 'longitude']]
df.to_csv(f'{EXPORT_PATH}/bixi_spatial_locations.csv')

# Used to replace station names in bike_station_departures.csv
station_names = df.index.to_list()


# Departures (Response Variable)
df = pd.read_csv(get_source_file_name('bike_station_departures'))
df.index = (
    df['day']
    .apply(lambda x: datetime.strptime(f'{2019}-{x}', '%Y-%j').strftime('%Y-%m-%d'))
    .to_list()
)
df = df.drop(columns=date_columns)
df = df[1:197]
df = df.transpose()
q_90 = np.quantile(df.to_numpy(na_value=0), 0.9)
df = df.applymap(lambda x: np.NaN if x > q_90 else x)
df[['2019-05-26', '2019-08-25', '2019-09-27']] = np.NaN
df.index = station_names
df.index.name = 'location'
df.to_csv(f'{EXPORT_PATH}/bixi_station_departures.csv')

# # Min Max scaling (0 - 1)
# df = (df - df.min()) / (df.max() - df.min())

# max_val = np.max(df.to_numpy(na_value=0))
# df = df / max_val



from __future__ import annotations

import math
from typing import TYPE_CHECKING, Any, Literal

import pandas as pd
import torch

from pyBKTR.tensor_ops import TSR

if TYPE_CHECKING:
    from pyBKTR.kernels import Kernel


def reshape_covariate_dfs(
    spatial_df: pd.DataFrame,
    temporal_df: pd.DataFrame,
    y_df: pd.DataFrame,
    y_column_name: str = 'y',
) -> pd.DataFrame:
    """
    Function used to transform covariates coming from two dataframes one for spatial and
    one for temporal into a single dataframe with the right shape for the BKTR Regressor.
    This is useful when the temporal covariates do not vary trough space and the spatial
    covariates do not vary trough time (Like in the BIXI example). The function also adds
    a column for the target variable at the beginning of the dataframe.

    Args:
        spatial_df (pd.DataFrame): Spatial covariates dataframe with an index named
          location and a shape of (n_locations, n_spatial_covariates)
        temporal_df (pd.DataFrame): Temporal covariates dataframe with an index named
            time and a shape of (n_times, n_temporal_covariates)
        y_df (pd.DataFrame): The dataframe containing the target variable. It should have
            a shape of (n_locations, n_times). The columns and index names of this dataframe
            should be correspond to the one of the spatial_df and temporal_df.

    Raises:
        ValueError: If the index names of spatial_df and y_df are not `location`
        ValueError: If the index name of temporal_df is not `time`
        ValueError: If the index values of spatial_df and y_df are not the same
        ValueError: If the temporal_df index values and y_df columns names are not the same

    Returns:
        pd.DataFrame: The reshaped covariates dataframe with a shape of
            (n_locations * n_times, 1 + n_spatial_covariates + n_temporal_covariates).
            The first column is the target variable and the other columns are the
            covariates. The index is a MultiIndex with the cartesian product of all
            location and time as levels.
    """
    spa_index_name = 'location'
    temp_index_name = 'time'
    if spatial_df.index.name != spa_index_name or y_df.index.name != spa_index_name:
        raise ValueError(f'Index names of spatial_df and y_df must be {spa_index_name}')
    if temporal_df.index.name != temp_index_name:
        raise ValueError(f'Index name of temporal_df must be {temp_index_name}')

    spatial_df_cp = spatial_df.copy()
    temporal_df_cp = temporal_df.copy()
    y_df_cp = y_df.copy()

    # Sort indexes and columns
    for df in [spatial_df_cp, temporal_df_cp, y_df_cp]:
        df.sort_index(inplace=True)
    y_df_cp.sort_index(axis=1, inplace=True)

    # Compare indexes values
    if spatial_df_cp.index.to_list() != y_df_cp.index.to_list():
        raise ValueError('Index values of spatial_df and y_df must be the same')
    if temporal_df_cp.index.to_list() != y_df_cp.columns.to_list():
        raise ValueError('temporal_df index values and y_df columns names must be the same')

    # Create cartesian product for covariates
    spatial_df_cp.reset_index(inplace=True)
    temporal_df_cp.reset_index(inplace=True)
    data_df = pd.merge(spatial_df_cp, temporal_df_cp, how='cross')
    data_df.set_index([spa_index_name, temp_index_name], inplace=True)
    data_df.sort_index(inplace=True)
    data_df.insert(0, y_column_name, y_df_cp.to_numpy().flatten())
    return data_df


def simulate_spatiotemporal_data(
    nb_locations: int,
    nb_time_points: int,
    nb_spatial_dimensions: int,
    spatial_scale: float,
    time_scale: float,
    spatial_covariates_means: list[float],
    temporal_covariates_means: list[float],
    spatial_kernel: Kernel,
    temporal_kernel: Kernel,
    noise_variance_scale: float,
) -> dict[str, pd.DataFrame]:
    """Simulate spatiotemporal data using kernel covariances.

    Args:
        nb_spatial_locations (int): Number of spatial locations
        nb_time_points (int): Number of time points
        nb_spatial_dimensions (int): Number of spatial dimensions
        spatial_scale (float): Spatial scale
        time_scale (float): Time scale
        spatial_covariates_means (list[float]): Spatial covariates means
        temporal_covariates_means (list[float]): Temporal covariates means
        spatial_kernel (Kernel): Spatial kernel
        temporal_kernel (Kernel): Temporal kernel
        noise_variance_scale (float): Noise variance scale

    Returns:
        dict[str, pd.DataFrame]: A dictionnary containing 4 dataframes:
            - `data_df` contains the response variable and the covariates
            - `spatial_positions_df` contains the spatial locations and their coordinates
            - `temporal_positions_df` contains the time points and their coordinates
            - `beta_df` contains the true beta coefficients
    """
    # Saving last fp_type to restore it at the end of the function
    # Using float64 to avoid numerical errors in simulation
    ini_fp_type = TSR.fp_type
    TSR.set_params(fp_type='float64')

    spa_pos = TSR.rand([nb_locations, nb_spatial_dimensions]) * spatial_scale
    temp_pos = TSR.tensor(
        [time_scale / (nb_time_points - 1) * x for x in range(0, nb_time_points)]
    )
    temp_pos = temp_pos.reshape([nb_time_points, 1])

    # Dimension labels
    s_dims = _get_dim_labels('s_dim', nb_spatial_dimensions)
    s_locs = _get_dim_labels('s_loc', nb_locations)
    t_points = _get_dim_labels('t_point', nb_time_points)
    s_covs = _get_dim_labels('s_cov', len(spatial_covariates_means))
    t_covs = _get_dim_labels('t_cov', len(temporal_covariates_means))

    spa_pos_df = pd.DataFrame(
        spa_pos.cpu(), columns=s_dims, index=pd.Index(s_locs, name='location')
    )
    temp_pos_df = pd.DataFrame(
        temp_pos.cpu(), columns=['time_val'], index=pd.Index(t_points, name='time')
    )

    spa_means = TSR.tensor(spatial_covariates_means)
    nb_spa_covariates = len(spa_means)
    spa_covariates = TSR.randn([nb_locations, nb_spa_covariates])
    spa_covariates = spa_covariates + spa_means

    temp_means = TSR.tensor(temporal_covariates_means)
    nb_temp_covariates = len(temp_means)
    temp_covariates = TSR.randn([nb_time_points, nb_temp_covariates])
    temp_covariates = temp_covariates + temp_means
    intercept_covariates = TSR.ones([nb_locations, nb_time_points, 1])
    covs = torch.concat(
        [
            intercept_covariates,
            spa_covariates.unsqueeze(1).expand(nb_locations, nb_time_points, nb_spa_covariates),
            temp_covariates.unsqueeze(0).expand(nb_locations, nb_time_points, nb_temp_covariates),
        ],
        dim=2,
    )
    nb_covs = 1 + nb_spa_covariates + nb_temp_covariates

    covs_covariance: torch.Tensor = torch.distributions.Wishart(
        df=nb_covs,
        covariance_matrix=TSR.eye(nb_covs),
    ).sample()

    spatial_kernel.set_positions(spa_pos_df)
    spatial_covariance = spatial_kernel.kernel_gen()
    temporal_kernel.set_positions(temp_pos_df)
    temporal_covariance = temporal_kernel.kernel_gen()

    # Use Matrix Normal distribution to sample beta values (to reduce memory usage)
    # the second covariance matrix is the Kronecker product of temporal and covariates covariances
    chol_spa: torch.Tensor = torch.linalg.cholesky(spatial_covariance)
    chol_temp_covs = torch.linalg.cholesky(
        TSR.kronecker_prod(temporal_covariance, covs_covariance)
    )
    temp_vals = TSR.randn([nb_locations, nb_time_points * nb_covs])
    temp_errs = TSR.randn([nb_locations, nb_time_points])
    beta_values = (chol_spa @ temp_vals @ chol_temp_covs.T).reshape(
        [nb_locations, nb_time_points, nb_covs]
    )
    y_val = torch.einsum('ijk,ijk->ij', covs, beta_values)
    err = temp_errs * (noise_variance_scale**0.5)
    y_val += err
    y_val = y_val.reshape([nb_locations * nb_time_points, 1])
    # We remove the intercept from the covariates
    covs = covs.reshape([nb_locations * nb_time_points, nb_covs])[:, 1:]
    spa_temp_df_index = pd.MultiIndex.from_product(
        [spa_pos_df.index, temp_pos_df.index],
        names=['location', 'time'],
    )
    data_df = pd.DataFrame(
        torch.concat([y_val, covs], dim=1).cpu(),
        columns=['y'] + s_covs + t_covs,
        index=spa_temp_df_index,
    )
    beta_df = pd.DataFrame(
        beta_values.reshape([nb_locations * nb_time_points, nb_covs]).cpu(),
        columns=['Intercept'] + s_covs + t_covs,
        index=spa_temp_df_index,
    )
    TSR.set_params(fp_type=ini_fp_type)
    return {
        'data_df': data_df,
        'spatial_positions_df': spa_pos_df,
        'temporal_positions_df': temp_pos_df,
        'beta_df': beta_df,
    }


def _get_dim_labels(dim_prefix: str, max_value: int) -> str:
    """Utility function to get the dimension labels for a
        given dimension prefix and max value.

    Args:
        dim_prefix (str): The prefix of the dimension labels
        max_value (int): The maximum value of the dimension labels

    Returns:
        str: The dimension labels
    """
    max_digits = len(str(max_value - 1))
    int_format = f'{{:0{max_digits}d}}'
    return [f'{dim_prefix}_{int_format.format(i)}' for i in range(max_value)]


def get_label_index_or_raise(
    label: Any, label_list: list[Any], label_type: Literal['spatial', 'temporal', 'feature']
) -> int:
    """Get the index of a label in a list of labels. If the label is not in the list,
    raise an error.

    Args:
        label (Any): The label for which we want to get the index
        label_list (List[Any]): The list of labels
        label_type (Literal['spatial', 'temporal', 'feature']): The label type

    Returns:
        int: The index of the label in the list

    :meta private:
    """
    try:
        return label_list.index(label)
    except ValueError:
        raise ValueError(f'Label `{label}` does not exist in {label_type} labels.')


class log(float):
    """
    Log wrapper for math.log to get better documentation

    :meta private:
    """

    def __new__(cls, log_val, *args, **kwargs):
        return super(log, cls).__new__(cls, math.log(log_val))

    def __repr__(cls):
        return f'Log({round(math.exp(cls), 10)})'



from typing import Any, Literal

import torch


class TSR:
    """
    Class containing all tensor operations used in BKTR. It serves also as a factory
    for all new tensors (giving the right type and device to each of them).
    """

    fp_type: Literal['float64', 'float32'] = 'float64'
    _dtype: torch.TensorType = torch.float64
    fp_device: str = 'cpu'
    _device: str = 'cpu'

    @classmethod
    @property
    def default_jitter(cls) -> float:
        match cls._dtype:
            case torch.float64:
                return 1e-8
            case torch.float32:
                return 1e-4
            case _:
                raise ValueError('The dtype used by TSR has no default mapped jitter value')

    @classmethod
    def set_params(
        cls,
        fp_type: Literal['float32', 'float64'] | None = None,
        fp_device: str | None = None,
        seed: int | None = None,
    ):
        if fp_type is not None:
            match fp_type:
                case 'float32':
                    cls._dtype = torch.float32
                case 'float64':
                    cls._dtype = torch.float64
                case _:
                    raise ValueError('fp_type must be either `float32` or `float64`')
        if fp_device is not None:
            cls._device = fp_device
            cls.fp_device = fp_device
        if seed is not None:
            torch.manual_seed(seed)

    @classmethod
    def tensor(cls, tensor_data: Any) -> torch.Tensor:
        return torch.tensor(tensor_data, dtype=cls._dtype, device=cls._device)

    @classmethod
    def is_tensor(cls, tensor_data: Any) -> bool:
        return torch.is_tensor(tensor_data)

    @classmethod
    def eye(cls, n: int):
        return torch.eye(n, dtype=cls._dtype, device=cls._device)

    @classmethod
    def ones(cls, tsr_dim: int | tuple[int]):
        return torch.ones(tsr_dim, dtype=cls._dtype, device=cls._device)

    @classmethod
    def zeros(cls, tsr_dim: int | tuple[int]):
        return torch.zeros(tsr_dim, dtype=cls._dtype, device=cls._device)

    @classmethod
    def rand(cls, tsr_dim: int | tuple[int]):
        return torch.rand(tsr_dim, dtype=cls._dtype, device=cls._device)

    @classmethod
    def randn(cls, tsr_dim: tuple[int]) -> torch.Tensor:
        return torch.randn(tsr_dim, dtype=cls._dtype, device=cls._device)

    @classmethod
    def randn_like(cls, input_tensor: torch.Tensor):
        return torch.randn_like(input_tensor, dtype=cls._dtype, device=cls._device)

    @classmethod
    def arange(cls, start: int, end: int, step: int = 1):
        return torch.arange(start, end, step, dtype=cls._dtype, device=cls._device)

    @classmethod
    def rand_choice(cls, choices_tsr: torch.Tensor, nb_sample: int, use_replace: bool = False):
        choices_indx = torch.multinomial(choices_tsr, nb_sample, replacement=use_replace)
        return choices_tsr[choices_indx]

    @staticmethod
    def kronecker_prod(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return torch.kron(a, b)

    @staticmethod
    def khatri_rao_prod(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        a_cols, b_cols = a.shape[1], b.shape[1]
        if a_cols != b_cols:
            raise ValueError(
                'Matrices must have the same number of columns to perform'
                f'khatri rao product, got {a_cols} and {b_cols}'
            )
        return torch.reshape(torch.einsum('ac,bc->abc', [a, b]), [-1, a.shape[1]])



import torch

from pyBKTR.tensor_ops import TSR


def check_dist_tensor_dimensions(
    x1: torch.Tensor,
    x2: torch.Tensor,
    expected_nb_dim: int = 2,
    expected_last_dim_shape: int | None = None,
):
    """Check that two tensors have valid dimensions for distance computation."""
    if not (TSR.is_tensor(x1) and TSR.is_tensor(x2)):
        raise ValueError('Distance params must be tensors')
    if not (x1.ndim == x2.ndim and x2.ndim == expected_nb_dim):
        raise ValueError(f'Distance params should have {expected_nb_dim} dimension(s)')
    if expected_last_dim_shape is not None and not (
        expected_last_dim_shape == x1.shape[-1] == x2.shape[-1]
    ):
        raise ValueError(
            f'Distance params last dimension should contain {expected_last_dim_shape} elements'
        )


def get_euclidean_dist_tsr(x: torch.Tensor) -> torch.Tensor:
    """Function to compute the euclidean distance between a tensor and its transpose."""
    check_dist_tensor_dimensions(x, x)
    xu1, xu2 = x.unsqueeze(1), x.unsqueeze(1).transpose(0, 1)
    return (xu1 - xu2).pow(2).sum(2).sqrt()


class GeoMercatorProjector:
    """Class to project coordinates with mercator projection on a 2D plane
    Project coordinates with mercator projection on a 2D plane for
    a given scale. Keep track of the scale and the center of the projection to
    be able to project new coordinates which is useful during interpolation.
    """

    ini_df = None
    x_mid_point = None
    y_mid_point = None
    coords_scale = None
    scale = None
    scaled_ini_df = None
    EARTH_RADIUM_KM = 6371

    def __init__(self, df, scale=10.0):
        self.ini_df = df
        km_df = self._km_from_coords_df(df)
        lon_x = km_df['lon_x']
        lat_y = km_df['lat_y']
        x_min, x_max = min(lon_x), max(lon_x)
        y_max, y_min = max(lat_y), min(lat_y)
        self.x_mid_point = (x_min + x_max) / 2
        self.y_mid_point = (y_min + y_max) / 2
        self.coords_scale = max(x_max - x_min, y_max - y_min)
        self.scale = scale
        self.scaled_ini_df = self._scale_and_center_df(km_df)

    def project_new_coords(self, df):
        km_df = self._km_from_coords_df(df)
        return self._scale_and_center_df(km_df)

    def _scale_and_center_df(self, df):
        new_df = df.copy()
        scaling_factor = self.scale / self.coords_scale
        new_df['lon_x'] = (new_df['lon_x'] - self.x_mid_point) * scaling_factor
        new_df['lat_y'] = (new_df['lat_y'] - self.y_mid_point) * scaling_factor
        return new_df

    def _km_from_coords_df(self, df):
        if not ('latitude' in df.columns and 'longitude' in df.columns):
            raise ValueError('Dataframe must have columns "latitude" and "longitude"')
        new_df = df.copy()
        lons = TSR.tensor(df['longitude'])
        lats = TSR.tensor(df['latitude'])
        x = (self.EARTH_RADIUM_KM / (2 * torch.pi)) * torch.deg2rad(lons)
        merc_n_y = torch.log(torch.tan(torch.pi / 4 + torch.deg2rad(lats) / 2))
        y = (self.EARTH_RADIUM_KM / (2 * torch.pi)) * merc_n_y
        new_df['lon_x'] = x.cpu().numpy()
        new_df['lat_y'] = y.cpu().numpy()
        new_df = new_df.drop(columns=['latitude', 'longitude'])
        return new_df



import itertools
from copy import deepcopy
from typing import Any, Literal

import numpy as np
import pandas as pd
import torch
from formulaic import Formula, model_matrix
from formulaic.errors import FormulaicError

from pyBKTR._likelihood_evaluator import MarginalLikelihoodEvaluator
from pyBKTR._result_logger import ResultLogger
from pyBKTR._samplers import (
    KernelParamSampler,
    PrecisionMatrixSampler,
    TauSampler,
    get_cov_decomp_chol,
    sample_norm_multivariate,
)
from pyBKTR.distances import GeoMercatorProjector
from pyBKTR.kernels import Kernel, KernelMatern, KernelSE
from pyBKTR.tensor_ops import TSR


class BKTRRegressor:
    """Class encapsulating the BKTR regression elements

    A BKTRRegressor holds all the key elements to accomplish the MCMC sampling
    algorithm (**Algorithm 1** of the paper).
    """

    __slots__ = [
        'data_df',
        'y',
        'omega',
        'covariates',
        'covariates_dim',
        'tau',
        'spatial_decomp',
        'temporal_decomp',
        'covs_decomp',
        'result_logger',
        'has_completed_sampling',
        'spatial_kernel',
        'temporal_kernel',
        'spatial_positions_df',
        'temporal_positions_df',
        'spatial_params_sampler',
        'temporal_params_sampler',
        'tau_sampler',
        'precision_matrix_sampler',
        'spatial_ll_evaluator',
        'temporal_ll_evaluator',
        'rank_decomp',
        'burn_in_iter',
        'sampling_iter',
        'max_iter',
        'a_0',
        'b_0',
        'formula',
        'spatial_labels',
        'temporal_labels',
        'feature_labels',
        'geo_coords_projector',
    ]
    data_df: pd.DataFrame
    y: torch.Tensor
    omega: torch.Tensor
    covariates: torch.Tensor
    covariates_dim: dict[str, int]
    tau: float
    # Covariate decompositions (change during iter)
    spatial_decomp: torch.Tensor  # U
    temporal_decomp: torch.Tensor  # V
    covs_decomp: torch.Tensor  # C or W
    # Kernels
    spatial_kernel: Kernel
    temporal_kernel: Kernel
    spatial_positions_df: pd.DataFrame
    temporal_positions_df: pd.DataFrame
    # Result Logger
    result_logger: ResultLogger
    has_completed_sampling: bool
    # Samplers
    spatial_params_sampler: KernelParamSampler
    temporal_params_sampler: KernelParamSampler
    tau_sampler: TauSampler
    precision_matrix_sampler: PrecisionMatrixSampler
    # Likelihood evaluators
    spatial_ll_evaluator: MarginalLikelihoodEvaluator
    temporal_ll_evaluator: MarginalLikelihoodEvaluator
    # Params
    rank_decomp: int
    burn_in_iter: int
    sampling_iter: int
    max_iter: int
    a_0: float
    b_0: float
    # Labels
    spatial_labels: list
    temporal_labels: list
    feature_labels: list
    # Geographical coordinates projector
    geo_coords_projector: GeoMercatorProjector | None

    # Constant string needed for dataframe index names
    spatial_index_name = 'location'
    temporal_index_name = 'time'

    def __init__(
        self,
        data_df: pd.DataFrame,
        spatial_positions_df: pd.DataFrame,
        temporal_positions_df: pd.DataFrame,
        rank_decomp: int = 10,
        burn_in_iter: int = 500,
        sampling_iter: int = 500,
        formula: None | str = None,
        spatial_kernel: Kernel = KernelMatern(smoothness_factor=3),
        temporal_kernel: Kernel = KernelSE(),
        sigma_r: float = 1e-2,
        a_0: float = 1e-6,
        b_0: float = 1e-6,
        has_geo_coords: bool = True,
        geo_coords_scale: float = 10,
    ):
        """Create a new *BKTRRegressor* object.

        Args:
            data_df (pd.DataFrame):  A dataframe containing all the covariates
                through time and space. It is important that the dataframe has a two
                indexes named `location` and `time` respectively. The dataframe should
                also contain every possible combinations of `location` and `time`
                (i.e. even missing rows should be filled present but filled with NaN).
                So if the dataframe has 10 locations and 5 time points, it should have
                50 rows (10 x 5). If formula is None, the dataframe should contain
                the response variable `Y` as the first column. Note that the covariate
                columns cannot contain NaN values, but the response variable can.
            formula (str | None, optional): A Wilkinson formula string to specify the relation
                between the response variable `Y` and the covariates (compatible with the
                Formulaic package).  If None, the first column of the data frame will be
                used as the response variable and all the other columns will be used as the
                covariates.  Defaults to None.
            rank_decomp (int): Rank of the CP decomposition (Paper -- :math:`R`). Defaults to 10.
            burn_in_iter (int): Iteration before sampling (Paper -- :math:`K_1`). Defaults to 500.
            sampling_iter (int): Sampling iteration number (Paper -- :math:`K_2`). Defaults to 500.
            spatial_positions_df (pd.DataFrame): Spatial kernel input
                tensor used to calculate covariates' distance. Vector of length equal to
                the number of location points.
            temporal_positions_df (pd.DataFrame): Temporal kernel input tensor
                used to calculate covariate distance. Vector of length equal to
                the number of time points.
            spatial_kernel (Kernel, optional): Spatial kernel Used.
                Defaults to KernelMatern(smoothness_factor=3).
            temporal_kernel (Kernel, optional): Temporal kernel used.
                Defaults to KernelSE().
            sigma_r (float, optional): Variance of the white noise process
                (Paper -- :math:`\\tau^{-1}`). Defaults to 1e-2.
            a_0 (float, optional): Initial value for the shape (:math:`\\alpha`) in the gamma
                function generating tau. Defaults to 1e-6.
            b_0 (float, optional): Initial value for the rate (:math:`\\beta`) in the gamma
                function generating tau. Defaults to 1e-6.
            has_geo_coords (bool, optional): Whether the spatial positions df use geographic
                coordinates (latitude, longitude). Defaults to True.
            geo_coords_scale (float, optional): Scale factor to convert geographic coordinates to
                euclidean 2D space via Mercator projection using x & y domains of
                [-scale/2, +scale/2]. Only used if has_geo_coords is TRUE. Defaults to 10.
        """
        self.has_completed_sampling = False
        self._verify_input_labels(data_df, spatial_positions_df, temporal_positions_df)

        # Sort all df indexes
        for df in [data_df, spatial_positions_df, temporal_positions_df]:
            df.sort_index(inplace=True)

        self.data_df = data_df
        self.temporal_positions_df = temporal_positions_df
        if has_geo_coords:
            self.geo_coords_projector = GeoMercatorProjector(
                spatial_positions_df, geo_coords_scale
            )
            self.spatial_positions_df = self.geo_coords_projector.scaled_ini_df
        else:
            self.spatial_positions_df = spatial_positions_df

        # Set formula and get model's matrix
        y_df, x_df = self._get_x_and_y_dfs_from_formula(data_df, formula)

        # Set labels
        self.spatial_labels = (
            y_df.index.get_level_values(self.spatial_index_name).unique().to_list()
        )
        self.temporal_labels = (
            y_df.index.get_level_values(self.temporal_index_name).unique().to_list()
        )
        self.feature_labels = x_df.columns.to_list()
        # Tensor assignation
        y_arr = y_df.to_numpy().reshape(len(self.spatial_labels), len(self.temporal_labels))
        self.omega = TSR.tensor(1 - np.isnan(y_arr))
        self.y = TSR.tensor(np.where(np.isnan(y_arr), 0, y_arr))
        covariates = TSR.tensor(x_df.to_numpy())
        self.tau = 1 / TSR.tensor(sigma_r)

        # Param assignation
        self.rank_decomp = rank_decomp
        self.burn_in_iter = burn_in_iter
        self.sampling_iter = sampling_iter
        self.max_iter = self.burn_in_iter + self.sampling_iter
        self.a_0 = a_0
        self.b_0 = b_0

        # Reshape covariates
        self._reshape_covariates(covariates, len(self.spatial_labels), len(self.temporal_labels))

        # Kernel assignation
        self.spatial_kernel = spatial_kernel
        self.temporal_kernel = temporal_kernel
        self.spatial_kernel.set_positions(self.spatial_positions_df)
        self.temporal_kernel.set_positions(self.temporal_positions_df)
        # Create first kernels
        self.spatial_kernel.kernel_gen()
        self.temporal_kernel.kernel_gen()

    def mcmc_sampling(self) -> dict[str, float | torch.Tensor]:
        """Launch the MCMC sampling process for a predefined number of iterations

        1. Sample spatial kernel hyperparameters
        2. Sample temporal kernel hyperparameters
        3. Sample the precision matrix from a wishart distribution
        4. Sample a new spatial covariate decomposition
        5. Sample a new feature covariate decomposition
        6. Sample a new temporal covariate decomposition
        7. Calculate respective errors for the iterations
        8. Sample a new tau value
        9. Collect all the important data for the iteration

        Returns:
            dict [str, float | torch.Tensor]: A dictionary with the MCMC sampling's results
        """
        self._initialize_params()
        for i in range(1, self.max_iter + 1):
            self._sample_kernel_hparam()
            self._sample_precision_wish()
            self._sample_spatial_decomp()
            self._sample_covariate_decomp()
            self._sample_temporal_decomp()
            self._set_errors_and_sample_precision_tau(i)
            self._collect_iter_values(i)
        self._log_final_iter_results()

    def predict(
        self,
        new_data_df: pd.DataFrame,
        new_spatial_positions_df: pd.DataFrame | None = None,
        new_temporal_positions_df: pd.DataFrame | None = None,
        jitter=1e-5,
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Predict the beta coefficients and response values for new data.

        Args:
            new_data_df (pd.DataFrame): New covariates. Must have the same columns as
                the covariates used to fit the model. The index should contain the combination
                of all old spatial coordinates with all new temporal coordinates, the combination
                of all new spatial coordinates with all old temporal coordinates, and the
                combination of all new spatial coordinates with all new temporal coordinates.
            new_spatial_positions_df (pd.DataFrame | None, optional): New spatial coordinates.
                If None, there should be no new spatial covariates. Defaults to None.
            new_temporal_positions_df (pd.DataFrame | None, optional): New temporal coordinates.
                If None, there should be no new temporal covariates. Defaults to None.
            jitter (float, optional): Jitter value to add to the diagonal of the precision matrix.
                Defaults to 1e-5.

        Returns:
            tuple[pd.DataFrame, pd.DataFrame]: A tuple of two dataframes. The first
            represents the beta forecasted for all new spatial locations or temporal points.
            The second represents the forecasted response for all new spatial locations or
            temporal points.
        """
        ini_fp_type = TSR.fp_type
        TSR.set_params(fp_type='float64')

        self._pred_valid_and_sort_data(
            new_data_df, new_spatial_positions_df, new_temporal_positions_df
        )

        spatial_positions_df = (
            pd.concat([self.spatial_positions_df, new_spatial_positions_df], axis=0)
            if new_spatial_positions_df is not None
            else self.spatial_positions_df
        )
        temporal_positions_df = (
            pd.concat([self.temporal_positions_df, new_temporal_positions_df], axis=0)
            if new_temporal_positions_df is not None
            else self.temporal_positions_df
        )
        data_df = pd.concat([self.data_df, new_data_df], axis=0)
        data_df_index_tups = list(
            itertools.product(
                spatial_positions_df.index.to_list(), temporal_positions_df.index.to_list()
            )
        )
        data_df = data_df.loc[data_df_index_tups, :]
        self._verify_input_labels(
            data_df,
            spatial_positions_df=spatial_positions_df,
            temporal_positions_df=temporal_positions_df,
        )
        all_betas = TSR.zeros(
            [
                len(spatial_positions_df),
                len(temporal_positions_df),
                len(self.formula.rhs),
                self.sampling_iter,
            ]
        )
        for i in range(self.sampling_iter):
            new_spa_decomp = self._pred_simu_new_decomp(
                'spatial', i, spatial_positions_df, new_spatial_positions_df, jitter
            )
            new_temp_decomp = self._pred_simu_new_decomp(
                'temporal', i, temporal_positions_df, new_temporal_positions_df, jitter
            )
            covs_decomp = self.result_logger.covs_decomp_per_iter[:, :, i]
            all_betas[:, :, :, i] = torch.einsum(
                'il,jl,kl->ijk', [new_spa_decomp, new_temp_decomp, covs_decomp]
            )

        new_betas = all_betas.mean(dim=-1)
        _, x_df = self._get_x_and_y_dfs_from_formula(data_df, self.formula)
        covariates = TSR.tensor(x_df.to_numpy()).reshape(
            [len(spatial_positions_df), len(temporal_positions_df), -1]
        )
        new_y_est = torch.einsum('ijk,ijk->ij', [new_betas, covariates])
        data_df_index = pd.MultiIndex.from_tuples(data_df_index_tups, names=['location', 'time'])
        new_beta_df = pd.DataFrame(
            new_betas.flatten(0, 1).cpu().numpy(),
            index=data_df_index,
            columns=x_df.columns,
        )
        new_y_df = pd.DataFrame(
            new_y_est.flatten().cpu().numpy(),
            index=data_df_index,
            columns=['y'],
        )
        new_locs = new_spatial_positions_df.index.get_level_values(0).unique()
        new_times = new_temporal_positions_df.index.get_level_values(0).unique()
        new_beta_df = new_beta_df[
            (new_beta_df.index.get_level_values(0).isin(new_locs))
            | (new_beta_df.index.get_level_values(1).isin(new_times))
        ]
        new_y_df = new_y_df[
            (new_y_df.index.get_level_values(0).isin(new_locs))
            | (new_y_df.index.get_level_values(1).isin(new_times))
        ]
        TSR.set_params(fp_type=ini_fp_type)
        return new_y_df, new_beta_df

    @property
    def summary(self) -> str:
        """Returns a summary of the MCMC regressor results

        Raises:
            RuntimeError: If the MCMC sampling has not been run yet

        Returns:
            str: A summary of the MCMC regressor results containing information about the
                MCMC sampling process and the estimated model's parameters.
        """
        if not self.has_completed_sampling:
            raise RuntimeError('Summary can only be accessed after MCMC sampling.')
        return self.result_logger.summary()

    @property
    def beta_covariates_summary_df(self) -> pd.DataFrame:
        if not self.has_completed_sampling:
            raise RuntimeError(
                'Covariate summary dataframe can only be accessed after MCMC sampling.'
            )
        return self.result_logger.beta_covariates_summary_df

    @property
    def y_estimates(self) -> pd.DataFrame:
        if not self.has_completed_sampling:
            raise RuntimeError('Y estimates can only be accessed after MCMC sampling.')
        y_est = self.result_logger.y_estimates_df.copy()
        y_est.iloc[:, 0].mask(self.omega.flatten().cpu() == 0, inplace=True)
        return y_est

    @property
    def imputed_y_estimates(self) -> pd.DataFrame:
        if not self.has_completed_sampling:
            raise RuntimeError('Imputed Y estimates can only be accessed after MCMC sampling.')
        return self.result_logger.y_estimates_df

    @property
    def beta_estimates(self) -> pd.DataFrame:
        if not self.has_completed_sampling:
            raise RuntimeError('Beta estimates can only be accessed after MCMC sampling.')
        return self.result_logger.beta_estimates_df

    def get_iterations_betas(
        self, spatial_label: Any, temporal_label: Any, feature_label: Any
    ) -> list[float]:
        """Return all sampled betas through sampling iterations for a given set of spatial,
            temporal and feature labels. Useful for plotting the distribution
            of sampled beta values.

        Args:
            spatial_label (Any): The spatial label for which we want to get the betas
            temporal_label (Any): The temporal label for which we want to get the betas
            feature_label (Any): The feature label for which we want to get the betas

        Returns:
            list[float]: The sampled betas through iteration for the given labels
        """
        if not self.has_completed_sampling:
            raise RuntimeError('Beta values can only be accessed after MCMC sampling.')
        beta_per_iter_tensor = self.result_logger.get_iteration_betas_tensor(
            [spatial_label], [temporal_label], [feature_label]
        )[0]
        return list(beta_per_iter_tensor.numpy())

    def get_beta_summary_df(
        self,
        spatial_labels: list[Any] = None,
        temporal_labels: list[Any] = None,
        feature_labels: list[Any] = None,
    ) -> pd.DataFrame:
        """Get a summary of estimated beta values. If no labels are given, then
        the summary is for all the betas. If labels are given, then the summary
        is for the given labels.

        Args:
            spatial_labels (list[Any], optional): The spatial labels to get the summary for.
                Defaults to None.
            temporal_labels (list[Any], optional): The temporal labels to get the summary for.
                Defaults to None.
            feature_labels (list[Any], optional): The feature labels to get the summary for.
                Defaults to None.

        Returns:
            pd.DataFrame: A dataframe with the summary for the given labels.
        """
        if not self.has_completed_sampling:
            raise RuntimeError('Beta values can only be accessed after MCMC sampling.')
        return self.result_logger.get_beta_summary_df(
            spatial_labels, temporal_labels, feature_labels
        )

    @property
    def hyperparameters_per_iter_df(self):
        if not self.has_completed_sampling:
            raise RuntimeError('Hyperparameters trace can only be accessed after MCMC sampling.')
        return self.result_logger.hyperparameters_per_iter_df

    @staticmethod
    def _verify_kernel_labels(
        kernel_positions: pd.DataFrame,
        expected_labels: set,
        kernel_type: Literal['spatial', 'temporal'],
    ):
        """Verify if kernel inputs are valid and align with covariates labels.

        Args:
            kernel_positions (pd.DataFrame): Kernel position vector to be provided (math:`x`).
            expected_labels (set): List of spatial/temporal labels used in covariates.
            kernel_type (Literal[&#39;spatial&#39;, &#39;temporal&#39;]): Type of kernel.

        Raises:
            ValueError: If kernel_positions size is not appropriate.
        """
        cov_related_indx_name = 'location' if kernel_type == 'spatial' else 'time'
        if kernel_positions.index.name != cov_related_indx_name:
            raise ValueError(
                f'`{kernel_type}_positions_df` must have a `{cov_related_indx_name}` index.'
            )
        if expected_labels != set(kernel_positions.index):
            raise ValueError(
                f'`{kernel_type}_positions_df` must contain in its {cov_related_indx_name}',
                f'index the unique values located in `data_df` {cov_related_indx_name} index.',
            )

    @classmethod
    def _verify_input_labels(
        cls,
        data_df: pd.DataFrame,
        spatial_positions_df: pd.DataFrame,
        temporal_positions_df: pd.DataFrame,
    ):
        """Verify the validity of BKTR dataframe inputs' labels

        Args:
            data_df (pd.DataFrame): The data_df from BKTRRegressor __init__
            spatial_positions_df (pd.DataFrame | None): spatial_positions_df from __init__
            temporal_positions_df (pd.DataFrame | None): temporal_positions_df from __init__

        Raises:
            ValueError: If data_df does not contain a multiindex with ['location', 'time']
            ValueError: If data_df location index do not correspond with spatial_positions_df
            ValueError: If data_df time index do not correspond with temporal_positions_df
        """
        if data_df.index.names != ['location', 'time']:
            raise ValueError('The data_df dataframe must have a [`location`, `time`] multiindex.')
        loc_set = set(data_df.index.get_level_values('location'))
        time_set = set(data_df.index.get_level_values('time'))
        product_set = set(itertools.product(loc_set, time_set))
        data_df_index_set = set(data_df.index)

        if data_df_index_set != product_set:
            raise ValueError(
                'The data_df dataframe must have a row for every possible'
                ' combination of location and time. Even if response values are missing (NaN).'
            )
        cls._verify_kernel_labels(spatial_positions_df, loc_set, 'spatial')
        cls._verify_kernel_labels(temporal_positions_df, time_set, 'temporal')

    def _get_x_and_y_dfs_from_formula(
        self, data_df: pd.DataFrame, formula: str | None
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Use formula to get x and y dataframes.

        Args:
            data_df (pd.DataFrame): The initial dataframe used to obtain the x and y dataframes.
            formula (str | None): Formula to give the y and X dataframes matrix. If formula is
                None, use the first column as y and all other columns as covariates.

        Raises:
            ValueError: The formula provided is not valid according to the formulaic package.
            ValueError: The formula provided does not contain 1 response variable.

        Returns:
            tuple[pd.DataFrame, pd.DataFrame]: The y and x dataframes respectively.
        """
        try:
            if formula is None:
                covariate_cols = [f'`{x}`' for x in data_df.columns[1:].to_list()]
                covariate_cols_str = ' + '.join(covariate_cols)
                self.formula = Formula(f'{data_df.columns[0]} ~ {covariate_cols_str}')
            else:
                self.formula = Formula(formula)

            y, x = model_matrix(self.formula, data_df, na_action='ignore')
        except FormulaicError as e:
            raise ValueError(
                'An error linked to the formula parameter occured.'
                ' Please check the formula syntax to comply with the formulaic package.'
                '\nFor more info see: https://matthewwardrop.github.io/formulaic/guides/'
                f'\nThe Formulaic error was:\n\t{e}'
            )
        y, x = pd.DataFrame(y), pd.DataFrame(x)
        if len(y.columns) != 1:
            raise ValueError(
                'The formula provided to the regressor is not valid.'
                ' It must contain one and only one response variable.'
            )
        return y, x

    def _reshape_covariates(
        self, covariate_tensor: torch.Tensor, nb_locations: int, nb_times: int
    ):
        """Reshape the covariate tensors into one single tensor and set it as a property

        Args:
            covariate_tensor (torch.Tensor): Tensor of covariates in a (M*N) x P shape
            nb_locations (int): The number of different locations (M)
            nb_times (int): The number of different times (N)
        """
        nb_covariates = covariate_tensor.shape[1]
        self.covariates_dim = {
            'nb_spaces': nb_locations,  # S
            'nb_times': nb_times,  # T
            'nb_covariates': nb_covariates,  # P
        }
        self.covariates = covariate_tensor.reshape([nb_locations, nb_times, nb_covariates])

    def _init_covariate_decomp(self):
        """Initialize CP decomposed covariate tensors with normally distributed random values"""
        rank_decomp = self.rank_decomp
        covs_dim = self.covariates_dim

        self.spatial_decomp = TSR.randn([covs_dim['nb_spaces'], rank_decomp])
        self.temporal_decomp = TSR.randn([covs_dim['nb_times'], rank_decomp])
        self.covs_decomp = TSR.randn([covs_dim['nb_covariates'], rank_decomp])

    def _create_result_logger(self):
        self.result_logger = ResultLogger(
            y=self.y,
            omega=self.omega,
            covariates=self.covariates,
            nb_burn_in_iter=self.burn_in_iter,
            nb_sampling_iter=self.sampling_iter,
            rank_decomp=self.rank_decomp,
            formula=self.formula,
            spatial_labels=self.spatial_labels,
            temporal_labels=self.temporal_labels,
            feature_labels=self.feature_labels,
            spatial_kernel=self.spatial_kernel,
            temporal_kernel=self.temporal_kernel,
        )

    def _create_likelihood_evaluators(self):
        """Create and set the evaluators for the spatial and the temporal likelihoods"""
        self.spatial_ll_evaluator = MarginalLikelihoodEvaluator(
            self.rank_decomp,
            self.covariates_dim['nb_covariates'],
            self.covariates,
            self.omega,
            self.y,
            is_transposed=False,
        )
        self.temporal_ll_evaluator = MarginalLikelihoodEvaluator(
            self.rank_decomp,
            self.covariates_dim['nb_covariates'],
            self.covariates,
            self.omega,
            self.y,
            is_transposed=True,
        )

    def _create_hparam_samplers(self):
        """Create hyperparameter samplers

        Create and set the hyperparameter samplers for the spatial/temporal kernel parameters,
        tau and the precision matrix
        """
        self.spatial_params_sampler = KernelParamSampler(
            kernel=self.spatial_kernel,
            marginal_ll_eval_fn=self._calc_spatial_marginal_ll,
        )
        self.temporal_params_sampler = KernelParamSampler(
            kernel=self.temporal_kernel,
            marginal_ll_eval_fn=self._calc_temporal_marginal_ll,
        )

        self.tau_sampler = TauSampler(self.a_0, self.b_0, self.omega.sum())

        self.precision_matrix_sampler = PrecisionMatrixSampler(
            self.covariates_dim['nb_covariates'], self.rank_decomp
        )

    def _calc_spatial_marginal_ll(self):
        """Calculate the spatial marginal likelihood"""
        return self.spatial_ll_evaluator.calc_likelihood(
            self.spatial_kernel.covariance_matrix, self.temporal_decomp, self.covs_decomp, self.tau
        )

    def _calc_temporal_marginal_ll(self):
        """Calculate the temporal marginal likelihood"""
        return self.temporal_ll_evaluator.calc_likelihood(
            self.temporal_kernel.covariance_matrix, self.spatial_decomp, self.covs_decomp, self.tau
        )

    def _sample_kernel_hparam(self):
        """Sample new kernel hyperparameters"""
        self.spatial_params_sampler.sample()
        self.temporal_params_sampler.sample()

    def _sample_precision_wish(self):
        """Sample the precision matrix from a Wishart distribution"""
        self.precision_matrix_sampler.sample(self.covs_decomp)

    def _sample_decomp_norm(
        self, initial_decomp: torch.Tensor, chol_l: torch.Tensor, uu: torch.Tensor
    ):
        """Sample a new covariate decomposition from a mulivariate normal distribution

        Args:
            initial_decomp (torch.Tensor): Decomposition of the previous iteration
            chol_l (torch.Tensor): The cholesky decomposition of the #TODO
            uu (torch.Tensor): _description_ #TODO

        Returns:
            torch.Tensor: A tensor containing the newly sampled covariate decomposition
        """
        precision_mat = chol_l.t()
        mean_vec = (
            self.tau
            * torch.linalg.solve_triangular(precision_mat, uu.unsqueeze(1), upper=True).squeeze()
        )
        return sample_norm_multivariate(mean_vec, precision_mat).reshape_as(initial_decomp.t()).t()

    def _sample_spatial_decomp(self):
        """Sample a new spatial covariate decomposition"""
        ll_eval = self.spatial_ll_evaluator
        self.spatial_decomp = self._sample_decomp_norm(
            self.spatial_decomp, ll_eval.chol_lu, ll_eval.uu
        )

    def _sample_covariate_decomp(self):
        """Sample a new covariate decomposition"""
        chol_res = get_cov_decomp_chol(
            self.spatial_decomp,
            self.temporal_decomp,
            self.covariates,
            self.rank_decomp,
            self.omega,
            self.tau,
            self.y,
            self.precision_matrix_sampler.wish_precision_tensor,
        )
        self.covs_decomp = self._sample_decomp_norm(
            self.covs_decomp, chol_res['chol_lc'], chol_res['cc']
        )

    def _sample_temporal_decomp(self):
        """Sample a new temporal covariate decomposition

        Need to recalculate uu and chol_u since covariate decomp changed
        """
        self._calc_temporal_marginal_ll()
        ll_eval = self.temporal_ll_evaluator
        self.temporal_decomp = self._sample_decomp_norm(
            self.temporal_decomp, ll_eval.chol_lu, ll_eval.uu
        )

    def _set_errors_and_sample_precision_tau(self, iter: int):
        """Sample a new tau and set errors"""
        self.result_logger.set_y_and_beta_estimates(self._decomposition_tensors, iter)
        self.result_logger.set_error_metrics()
        self.tau = self.tau_sampler.sample(self.result_logger.total_sq_error)

    @property
    def _decomposition_tensors(self):
        """Get a dict of current iteration decomposition needed to calculate estimated betas"""
        # TODO this could be arguments in the result logger
        return {
            'spatial_decomp': self.spatial_decomp,
            'temporal_decomp': self.temporal_decomp,
            'covs_decomp': self.covs_decomp,
        }

    def _collect_iter_values(self, iter: int):
        """Collect current iteration values inside the historical data tensor list"""
        self.result_logger.collect_iter_samples(iter, float(self.tau))

    def _log_final_iter_results(self):
        self.result_logger.log_final_iter_results()
        self.has_completed_sampling = True

    def _initialize_params(self):
        """Initialize all parameters that are needed before we start the MCMC sampling"""
        self._init_covariate_decomp()
        self._create_result_logger()
        self._create_likelihood_evaluators()
        self._create_hparam_samplers()
        # Calcultate first likelihoods
        self._calc_spatial_marginal_ll()
        self._calc_temporal_marginal_ll()

    def _pred_simu_new_decomp(
        self,
        pred_type: Literal['spatial', 'temporal'],
        iter_no: int,
        position_df: pd.DataFrame,
        new_position_df: pd.DataFrame | None,
        jitter: float | None,
    ) -> torch.Tensor:
        """Predict new decomposition values for a given iteration using the interpolation
            algorithm described in the paper.

        Args:
            pred_type (Literal[&#39;spatial&#39;, &#39;temporal&#39;]): Type of prediction.
            iter_no (int): Iteration number for which we want to predict new decomposition values.
            position_df (pd.DataFrame): The position dataframe containing the positions (locations
                or time points depending on the prediction type) used during the MCMC sampling.
            new_position_df (pd.DataFrame | None): The position dataframe containing the positions
            (locations or time points) for which we want to predict new decomposition values.
            jitter (float | None): The jitter value to use for the interpolation algorithm.

        Returns:
            torch.Tensor: A tensor containing the predicted decomposition values for the new
                positions.
        """
        old_decomp = (
            self.result_logger.spatial_decomp_per_iter[:, :, iter_no]
            if pred_type == 'spatial'
            else self.result_logger.temporal_decomp_per_iter[:, :, iter_no]
        )
        if new_position_df is None:
            return old_decomp
        nb_new_pos = len(new_position_df)
        old_kernel = self.spatial_kernel if pred_type == 'spatial' else self.temporal_kernel
        new_kernel = deepcopy(old_kernel)
        for param in new_kernel.parameters:
            if not param.is_fixed:
                param_full_repr = f'{pred_type.capitalize()} - {param.full_name}'
                param.value = self.result_logger.hyperparameters_per_iter_df.loc[
                    iter_no + 1, param_full_repr
                ]
        new_kernel.set_positions(position_df)
        cov_mat = new_kernel.kernel_gen()
        old_cov = cov_mat[:-nb_new_pos, :-nb_new_pos]
        new_old_cov = cov_mat[-nb_new_pos:, :-nb_new_pos]
        old_new_cov = cov_mat[:-nb_new_pos, -nb_new_pos:]
        new_cov = cov_mat[-nb_new_pos:, -nb_new_pos:]
        new_decomp_mus = new_old_cov @ old_cov.inverse() @ old_decomp
        new_decomp_cov = new_cov - (new_old_cov @ old_cov.inverse() @ old_new_cov)
        new_decomp_cov = (new_decomp_cov + new_decomp_cov.T) / 2
        if jitter is not None:
            new_decomp_cov += jitter * TSR.eye(new_decomp_cov.shape[0])
        new_decomp = (
            torch.distributions.MultivariateNormal(new_decomp_mus.T, new_decomp_cov).sample().T
        )
        return torch.concat([old_decomp, new_decomp], dim=0)

    @staticmethod
    def _check_pred_dfs_integrity(
        old_df_name: str, old_df: pd.DataFrame, new_df: pd.DataFrame
    ) -> None:
        """Check that the new dataframe has the same columns and index as the old one,
        that the new df index labels are unique and that the new df index labels are
        not a subset of the old df index labels.

        Args:
            old_df_name (str): Previous name for the dataframe use in regression.
            old_df (pd.DataFrame): Previous dataframe use in regression.
            new_df (pd.DataFrame): New dataframe use in for prediction.
        """
        if old_df.columns.to_list() != new_df.columns.to_list():
            raise ValueError(
                f'The `new_{old_df_name}` columns should correspond with `{old_df_name}` columns.'
            )
        new_pos_labels_set = set(new_df.index.unique().to_list())
        old_pos_labels_set = set(old_df.index.unique().to_list())
        if len(old_pos_labels_set) != len(old_df):
            raise ValueError(f'All index labels in new_{old_df_name} should be unique.')
        if new_pos_labels_set & old_pos_labels_set:
            raise ValueError(
                f'The index labels in new_{old_df_name} should not exists in {old_df_name}.'
            )

    def _pred_valid_and_sort_data(
        self,
        new_data_df: pd.DataFrame,
        new_spatial_positions_df: pd.DataFrame | None,
        new_temporal_positions_df: pd.DataFrame | None,
    ) -> None:
        """Check that the new data and covariates are valid and sort them by index.

        Args:
            new_data_df (pd.DataFrame): A dataframe with the new data to predict.
            new_spatial_positions_df (pd.DataFrame | None): A dataframe with new spatial locations.
            new_temporal_positions_df (pd.DataFrame | None): A dataframe with new time positions.
        """
        if new_spatial_positions_df is None and new_temporal_positions_df is None:
            raise ValueError(
                'At least one of new_positions_spatial_df and'
                ' new_positions_temporal_df must be provided.'
            )
        for df in [
            new_data_df,
            new_spatial_positions_df,
            new_temporal_positions_df,
        ]:
            if df is not None:
                df.sort_index(inplace=True)

        self._check_pred_dfs_integrity('data_df', self.data_df, new_data_df)
        data_df_labs_set = set(new_data_df.index.to_list())

        # Check spatial locations dataframes integrity
        x_spa_loc_labels = None
        new_spa_needed_indx = set()
        if new_spatial_positions_df is not None:
            self._check_pred_dfs_integrity(
                'spatial_positions_df', self.spatial_positions_df, new_spatial_positions_df
            )
            x_spa_loc_labels = new_spatial_positions_df.index.unique().to_list()
            old_time_labels = self.temporal_positions_df.index.to_list()
            new_spa_needed_indx = set(itertools.product(x_spa_loc_labels, old_time_labels))
            if not new_spa_needed_indx.issubset(data_df_labs_set):
                raise ValueError(
                    'The index of `new_data_df` should include the combination of all previous'
                    ' time points and all new locations provided in `new_spatial_positions_df`.'
                )

        # Check temporal points dataframes integrity
        x_temp_loc_labels = None
        new_temp_needed_indx = set()
        if new_temporal_positions_df is not None:
            self._check_pred_dfs_integrity(
                'temporal_positions_df', self.temporal_positions_df, new_temporal_positions_df
            )
            x_temp_loc_labels = new_temporal_positions_df.index.unique().to_list()
            old_spa_labels = self.spatial_positions_df.index.to_list()
            new_temp_needed_indx = set(itertools.product(old_spa_labels, x_temp_loc_labels))
            if not new_temp_needed_indx.issubset(data_df_labs_set):
                raise ValueError(
                    'The index of `new_data_df` should include the combination of all previous'
                    ' locations and all new time points provided in `new_temporal_positions_df`.'
                )

        # Check combination of new spatial and temporal dataframes integrity
        combi_needed_indx = set()
        if x_spa_loc_labels and x_temp_loc_labels:
            combi_needed_indx = set(itertools.product(x_spa_loc_labels, x_temp_loc_labels))
            if not combi_needed_indx.issubset(data_df_labs_set):
                raise ValueError(
                    'The index of`new_data_df` should include the combination of all new time'
                    'points in `new_temporal_positions_df` and all new locations'
                    ' in `new_spatial_positions_df`.'
                )

        if data_df_labs_set != new_spa_needed_indx | new_temp_needed_indx | combi_needed_indx:
            raise ValueError(
                'The index of `data_df` should only contain:'
                '\n\t1. The combination of all new locations and old time points;'
                '\n\t2. The combination of all new time points and old locations;'
                '\n\t3. The combination of all new time poins and new locations;'
            )
